"""Functions to convert curved Rhino geometries into planar ladybug ones."""
from .config import tolerance

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D
    from ladybug_geometry.geometry3d.face import Face3D
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry.\n{}".format(e))

try:
    import Rhino.Geometry as rg
except ImportError as e:
    raise ImportError(
        "Failed to import Rhino.\n{}".format(e))

import sys
if (sys.version_info > (3, 0)):  # python 3
    xrange = range


"""____________INDIVIDUAL SURFACES TO PLANAR____________"""


def planar_face_curved_edge_vertices(b_face, count, meshing_parameters):
    """Extract vertices from a planar brep face loop that has one or more curved edges.

    This method ensures vertices along the curved edge are generated in a way that
    they align with an extrusion of that edge. Alignment may not be possible when
    the adjoining curved surface is not an extrusion.

    Args:
        b_face: A brep face with the curved edge.
        count: An integer for the index of the loop to extract.
        meshing_parameters: Rhino Meshing Parameters to describe how
            curved edge should be converted into planar elements.

    Returns:
        A list of ladybug Point3D objects representing the input planar face.
    """
    loop_pcrv = b_face.Loops.Item[count].To3dCurve()
    f_norm = b_face.NormalAt(0, 0)
    if f_norm.Z < 0:
        loop_pcrv.Reverse()
    loop_verts = []
    try:
        loop_pcrvs = [loop_pcrv.SegmentCurve(i)
                      for i in xrange(loop_pcrv.SegmentCount)]
    except Exception:
        try:
            loop_pcrvs = [loop_pcrv[0]]
        except Exception:
            loop_pcrvs = [loop_pcrv]
    for seg in loop_pcrvs:
        if seg.Degree == 1:
            loop_verts.append(_point3d(seg.PointAtStart))
        else:
            # Ensure curve subdivisions align with adjacent curved faces
            seg_mesh = rg.Mesh.CreateFromSurface(
                rg.Surface.CreateExtrusion(seg, f_norm),
                meshing_parameters)
            for i in xrange(seg_mesh.Vertices.Count / 2 - 1):
                loop_verts.append(_point3d(seg_mesh.Vertices[i]))
    return loop_verts


def curved_surface_faces(b_face, meshing_parameters):
    """Extract Face3D objects from a curved brep face.

    Args:
        b_face: A curved brep face.
        meshing_parameters: Rhino Meshing Parameters to describe how
            curved edge should be converted into planar elements.

    Returns:
        A list of ladybug Face3D objects that together approximate the input
        curved surface.
    """
    if b_face.OrientationIsReversed:
        b_face.Reverse(0, True)
    face_brep = b_face.DuplicateFace(True)
    meshed_brep = rg.Mesh.CreateFromBrep(face_brep, meshing_parameters)[0]
    return mesh_faces_to_face3d(meshed_brep)


"""____________SOLID BREPS TO PLANAR____________"""


def curved_solid_faces(brep, meshing_parameters, ignore_sliver=True):
    """Extract Face3D objects from a curved solid brep.

    This method ensures that the resulting Face3Ds sill form a closed solid if
    the input brep is closed. This is accomplished by meshing the solid Brep
    altogether.

    Args:
        brep: A curved solid brep.
        meshing_parameters: Rhino Meshing Parameters to describe how curved surfaces
            should be converted into planar elements. If None, Rhino's Default
            Meshing Parameters will be used.
        ignore_sliver: A Boolean to note whether tiny sliver faces should simply be
            excluded from the output (True) or whether a None should be put in
            their place (False). The latter is useful when reporting to the user
            that certain tiny geometries interfered with the planarization
            process and the Rhino model tolerance should probably be lowered
            in order to get a better planar representation. (Default: True).

    Returns:
        A list of ladybug Face3D objects that together approximate the input brep.
    """
    # mesh the geometry as a solid
    mesh_par = meshing_parameters or rg.MeshingParameters.Default  # default
    meshed_geo = rg.Mesh.CreateFromBrep(brep, mesh_par)

    # evaluate each mesh face to see what larger brep face it is a part of
    faces = []
    for mesh, b_face in zip(meshed_geo, brep.Faces):
        if b_face.IsPlanar(tolerance):  # only take the naked vertices of planar faces
            naked_edges = mesh.GetNakedEdges()
            all_verts = []
            for loop_pline in naked_edges:  # each loop_pline is a boundary/hole
                all_verts.append([_point3d(loop_pline.Item[i])
                                  for i in xrange(loop_pline.Count - 1)])
            if len(all_verts) == 1:  # No holes in the shape
                faces.append(Face3D(all_verts[0]))
            else:  # There's at least one hole in the shape
                faces.append(
                    Face3D(boundary=all_verts[0], holes=all_verts[1:]))
        else:
            faces.extend(mesh_faces_to_face3d(mesh))
    # remove colinear vertices as the meshing process makes a lot of them
    final_faces = []
    for face in faces:
        try:
            final_faces.append(face.remove_colinear_vertices(tolerance))
        except AssertionError:  # tiny sliver Face that should not be included
            if not ignore_sliver:
                final_faces.append(None)
    return final_faces


"""________________EXTRA HELPER FUNCTIONS________________"""


def has_curved_face(brep):
    """Test if a Rhino Brep has a curved face.

    Args:
        brep: A Rhino Brep to test whether it has a curved face.
    """
    for b_face in brep.Faces:
        if not b_face.IsPlanar(tolerance):
            return True
    return False


def mesh_faces_to_face3d(mesh):
    """Convert a curved Rhino mesh faces into planar ladybug_geometry Face3D.

    Args:
        mesh: A curved Rhino mesh.

    Returns:
        A list of ladybug Face3D objects derived from the input mesh faces.
    """
    faces = []
    for m_face in mesh.Faces:
        if m_face.IsQuad:
            lb_face = Face3D(
                tuple(_point3d(mesh.Vertices[i]) for i in
                      (m_face.A, m_face.B, m_face.C, m_face.D)))
            if lb_face.check_planar(tolerance, False):
                faces.append(lb_face)
            else:
                lb_face_1 = Face3D(
                    tuple(_point3d(mesh.Vertices[i]) for i in
                          (m_face.A, m_face.B, m_face.C)))
                lb_face_2 = Face3D(
                    tuple(_point3d(mesh.Vertices[i]) for i in
                          (m_face.C, m_face.D, m_face.A)))
                faces.extend([lb_face_1, lb_face_2])
        else:
            lb_face = Face3D(
                tuple(_point3d(mesh.Vertices[i]) for i in
                      (m_face.A, m_face.B, m_face.C)))
            faces.append(lb_face)
    return faces


def _point3d(point):
    """Ladybug Point3D from Rhino Point3d."""
    return Point3D(point.X, point.Y, point.Z)



"""Functions to bake from Ladybug geometries into a Rhino document."""
from .fromgeometry import from_point2d, from_arc2d, from_polyline2d, from_mesh2d, \
    from_point3d, from_plane, from_arc3d, from_polyline3d, \
    from_mesh3d, from_face3d, from_polyface3d, from_sphere, from_cone, from_cylinder
from .color import color_to_color, gray

try:
    from System.Drawing import Color
except ImportError as e:
    raise ImportError("Failed to import Windows/.NET libraries\n{}".format(e))

try:
    import Rhino.Geometry as rg
    import Rhino.DocObjects as docobj
    import Rhino.RhinoDoc as rhdoc
    doc = rhdoc.ActiveDoc
except ImportError as e:
    raise ImportError("Failed to import Rhino document attributes.\n{}".format(e))

"""____________BAKE 2D GEOMETRY TO THE RHINO SCENE____________"""


def bake_vector2d(vector, z=0, layer_name=None, attributes=None):
    """Add ladybug Ray2D to the Rhino scene as a Line with an Arrowhead."""
    seg = (rg.Point3d(0, 0, z), rg.Point3d(vector.x, vector.y, z))
    attrib = _get_attributes(layer_name, attributes)
    attrib.ObjectDecoration = docobj.ObjectDecoration.EndArrowhead
    return doc.Objects.AddLine(seg[0], seg[1], attrib)


def bake_point2d(point, z=0, layer_name=None, attributes=None):
    """Add ladybug Point2D to the Rhino scene as a Point."""
    pt = from_point2d(point, z)
    return doc.Objects.AddPoint(pt, _get_attributes(layer_name, attributes))


def bake_ray2d(ray, z=0, layer_name=None, attributes=None):
    """Add ladybug Ray2D to the Rhino scene as a Line with an Arrowhead."""
    seg = (from_point2d(ray.p, z), from_point2d(ray.p + ray.v, z))
    attrib = _get_attributes(layer_name, attributes)
    attrib.ObjectDecoration = docobj.ObjectDecoration.EndArrowhead
    return doc.Objects.AddLine(seg[0], seg[1], attrib)


def bake_linesegment2d(line, z=0, layer_name=None, attributes=None):
    """Add ladybug LineSegment2D to the Rhino scene as a Line."""
    seg = (from_point2d(line.p1, z), from_point2d(line.p2, z))
    return doc.Objects.AddLine(seg[0], seg[1], _get_attributes(layer_name, attributes))


def bake_polygon2d(polygon, z=0, layer_name=None, attributes=None):
    """Add ladybug Polygon2D to the Rhino scene as a Polyline."""
    pgon = [from_point2d(pt, z) for pt in polygon.vertices] + \
        [from_point2d(polygon[0], z)]
    return doc.Objects.AddPolyline(pgon, _get_attributes(layer_name, attributes))


def bake_arc2d(arc, z=0, layer_name=None, attributes=None):
    """Add ladybug Arc2D to the Rhino scene as an Arc or a Circle."""
    rh_arc = from_arc2d(arc, z)
    if arc.is_circle:
        return doc.Objects.AddCircle(rh_arc, _get_attributes(layer_name, attributes))
    else:
        return doc.Objects.AddArc(rh_arc, _get_attributes(layer_name, attributes))


def bake_polyline2d(polyline, z=0, layer_name=None, attributes=None):
    """Add ladybug Polyline2D to the Rhino scene as a Curve."""
    rh_crv = from_polyline2d(polyline, z)
    return doc.Objects.AddCurve(rh_crv, _get_attributes(layer_name, attributes))


def bake_mesh2d(mesh, z=0, layer_name=None, attributes=None):
    """Add ladybug Mesh2D to the Rhino scene as a Mesh."""
    _mesh = from_mesh2d(mesh, z)
    return doc.Objects.AddMesh(_mesh, _get_attributes(layer_name, attributes))


"""____________BAKE 3D GEOMETRY TO THE RHINO SCENE____________"""


def bake_vector3d(vector, layer_name=None, attributes=None):
    """Add ladybug Ray2D to the Rhino scene as a Line with an Arrowhead."""
    seg = (rg.Point3d(0, 0, 0), rg.Point3d(vector.x, vector.y, vector.z))
    attrib = _get_attributes(layer_name, attributes)
    attrib.ObjectDecoration = docobj.ObjectDecoration.EndArrowhead
    return doc.Objects.AddLine(seg[0], seg[1], attrib)


def bake_point3d(point, layer_name=None, attributes=None):
    """Add ladybug Point3D to the Rhino scene as a Point."""
    pt = from_point3d(point)
    return doc.Objects.AddPoint(pt, _get_attributes(layer_name, attributes))


def bake_ray3d(ray, layer_name=None, attributes=None):
    """Add ladybug Ray2D to the Rhino scene as a Line with an Arrowhead."""
    seg = (from_point3d(ray.p), from_point3d(ray.p + ray.v))
    attrib = _get_attributes(layer_name, attributes)
    attrib.ObjectDecoration = docobj.ObjectDecoration.EndArrowhead
    return doc.Objects.AddLine(seg[0], seg[1], attrib)


def bake_plane(plane, layer_name=None, attributes=None):
    """Add ladybug Plane to the Rhino scene as a Rectangle."""
    rh_pln = from_plane(plane)
    r = 10  # default radius for a plane object in rhino model units
    interval = rg.Interval(-r / 2, r / 2)
    rect = rg.Rectangle3d(rh_pln, interval, interval)
    return doc.Objects.AddRectangle(rect, _get_attributes(layer_name, attributes))


def bake_linesegment3d(line, layer_name=None, attributes=None):
    """Add ladybug LineSegment3D to the Rhino scene as a Line."""
    seg = (from_point3d(line.p1), from_point3d(line.p2))
    return doc.Objects.AddLine(seg[0], seg[1], _get_attributes(layer_name, attributes))


def bake_arc3d(arc, layer_name=None, attributes=None):
    """Add ladybug Arc3D to the Rhino scene as an Arc or Circle."""
    rh_arc = from_arc3d(arc)
    if arc.is_circle:
        return doc.Objects.AddCircle(rh_arc, _get_attributes(layer_name, attributes))
    else:
        return doc.Objects.AddArc(rh_arc, _get_attributes(layer_name, attributes))


def bake_polyline3d(polyline, layer_name=None, attributes=None):
    """Add ladybug Polyline3D to the Rhino scene as a Curve."""
    rh_crv = from_polyline3d(polyline)
    return doc.Objects.AddCurve(rh_crv, _get_attributes(layer_name, attributes))


def bake_mesh3d(mesh, layer_name=None, attributes=None):
    """Add ladybug Mesh3D to the Rhino scene as a Mesh."""
    _mesh = from_mesh3d(mesh)
    return doc.Objects.AddMesh(_mesh, _get_attributes(layer_name, attributes))


def bake_face3d(face, layer_name=None, attributes=None):
    """Add ladybug Face3D to the Rhino scene as a Brep."""
    _face = from_face3d(face)
    return doc.Objects.AddBrep(_face, _get_attributes(layer_name, attributes))


def bake_polyface3d(polyface, layer_name=None, attributes=None):
    """Add ladybug Polyface3D to the Rhino scene as a Brep."""
    rh_polyface = from_polyface3d(polyface)
    return doc.Objects.AddBrep(rh_polyface, _get_attributes(layer_name, attributes))


def bake_sphere(sphere, layer_name=None, attributes=None):
    """Add ladybug Sphere to the Rhino scene as a Brep."""
    rh_sphere = from_sphere(sphere).ToBrep()
    return doc.Objects.AddBrep(rh_sphere, _get_attributes(layer_name, attributes))


def bake_cone(cone, layer_name=None, attributes=None):
    """Add ladybug Cone to the Rhino scene as a Brep."""
    rh_cone = from_cone(cone).ToBrep()
    return doc.Objects.AddBrep(rh_cone, _get_attributes(layer_name, attributes))


def bake_cylinder(cylinder, layer_name=None, attributes=None):
    """Add ladybug Cylinder to the Rhino scene as a Brep."""
    rh_cylinder = from_cylinder(cylinder).ToBrep()
    return doc.Objects.AddBrep(rh_cylinder, _get_attributes(layer_name, attributes))


"""________ADDITIONAL 3D GEOMETRY TRANSLATORS________"""


def bake_mesh3d_as_hatch(mesh, layer_name=None, attributes=None):
    """Add ladybug Mesh3D to the Rhino scene as a colored hatch."""
    # get a list of colors that align with the mesh faces
    if mesh.colors is not None:
        if mesh.is_color_by_face:
            colors = [color_to_color(col) for col in mesh.colors]
        else:  # compute the average color across the vertices
            colors, v_cols = [], mesh.colors
            for face in mesh.faces:
                red = int(sum(v_cols[f].r for f in face) / len(face))
                green = int(sum(v_cols[f].g for f in face) / len(face))
                blue = int(sum(v_cols[f].b for f in face) / len(face))
                colors.append(Color.FromArgb(255, red, green, blue))
    else:
        colors = [gray()] * len(mesh.faces)

    # create hatches from each of the faces and get an aligned list of colors
    hatches, hatch_colors = [], []
    vertices = mesh.vertices
    for face, color in zip(mesh.faces, colors):
        f_verts = [from_point3d(vertices[f]) for f in face]
        f_verts.append(f_verts[0])
        p_line = rg.PolylineCurve(f_verts)
        if p_line.IsPlanar():
            hatches.append(rg.Hatch.Create(p_line, 0, 0, 0)[0])
            hatch_colors.append(color)
        elif len(face) == 4:
            p_line_1 = rg.PolylineCurve(f_verts[:3] + [f_verts[0]])
            p_line_2 = rg.PolylineCurve(f_verts[-3:] + [f_verts[-3]])
            hatches.append(rg.Hatch.Create(p_line_1, 0, 0, 0)[0])
            hatches.append(rg.Hatch.Create(p_line_2, 0, 0, 0)[0])
            hatch_colors.extend((color, color))

    # bake the hatches into the scene
    guids = []
    for hatch, color in zip(hatches, hatch_colors):
        attribs = _get_attributes(layer_name, attributes)
        attribs.ColorSource = docobj.ObjectColorSource.ColorFromObject
        attribs.ObjectColor = color
        guids.append(doc.Objects.AddHatch(hatch, attribs))

    # group the hatches so that they are easy to handle in the Rhino scene
    group_t = doc.Groups
    docobj.Tables.GroupTable.Add(group_t, guids)
    return guids


"""________________EXTRA HELPER FUNCTIONS________________"""


def _get_attributes(layer_name=None, attributes=None):
    """Get Rhino object attributes."""
    attributes = doc.CreateDefaultAttributes() if attributes is None else attributes
    if layer_name is None:
        return attributes
    elif isinstance(layer_name, int):
        attributes.LayerIndex = layer_name
    elif layer_name is not None:
        attributes.LayerIndex = _get_layer(layer_name)
    return attributes


def _get_layer(layer_name):
    """Get a layer index from the Rhino document from the layer name."""
    layer_table = doc.Layers  # layer table
    layer_index = layer_table.FindByFullPath(layer_name, True)
    if layer_index < 0:
        all_layers = layer_name.split('::')
        parent_name = all_layers[0]
        layer_index = layer_table.FindByFullPath(parent_name, True)
        if layer_index < 0:
            parent_layer = docobj.Layer()
            parent_layer.Name = parent_name
            layer_index = layer_table.Add(parent_layer)
        for lay in all_layers[1:]:
            parent_name = '{}::{}'.format(parent_name, lay)
            parent_index = layer_index
            layer_index = layer_table.FindByFullPath(parent_name, True)
            if layer_index < 0:
                parent_layer = docobj.Layer()
                parent_layer.Name = lay
                parent_layer.ParentLayerId = layer_table[parent_index].Id
                layer_index = layer_table.Add(parent_layer)
    return layer_index



"""Functions to handle intersection of Rhino geometries.

These represent geometry computation methods  that are either not supported by
ladybug_geometry or there are much more efficient versions of them in Rhino.
"""
import math
import array as specializedarray

try:
    import System.Threading.Tasks as tasks
except ImportError as e:
    print('Failed to import Windows/.NET libraries\nParallel processing functionality '
          'will not be available\n{}'.format(e))

try:
    import Rhino.Geometry as rg
except ImportError as e:
    raise ImportError("Failed to import Rhino.\n{}".format(e))

from .config import tolerance, rhino_version


def join_geometry_to_mesh(geometry):
    """Convert an array of Rhino Breps and/or Meshes into a single Rhino Mesh.

    This is a typical pre-step before using the intersect_mesh_rays functions.

    Args:
        geometry: An array of Rhino Breps or Rhino Meshes.
    """
    joined_mesh = rg.Mesh()
    for geo in geometry:
        if isinstance(geo, rg.Brep):
            meshes = rg.Mesh.CreateFromBrep(geo, rg.MeshingParameters.Default)
            for mesh in meshes:
                joined_mesh.Append(mesh)
        elif isinstance(geo, rg.Mesh):
            joined_mesh.Append(geo)
        else:
            raise TypeError('Geometry must be either a Brep or a Mesh. '
                            'Not {}.'.format(type(geo)))
    return joined_mesh


def join_geometry_to_brep(geometry):
    """Convert an array of Rhino Breps and/or Meshes into a single Rhino Brep.

    This is a typical pre-step before using the ray tracing functions.

    Args:
        geometry: An array of Rhino Breps or Rhino Meshes.
    """
    joined_mesh = join_geometry_to_mesh(geometry)
    return rg.Brep.CreateFromMesh(joined_mesh, False)


def bounding_box(geometry, high_accuracy=False):
    """Get a Rhino bounding box around an input Rhino Mesh or Brep.

    This is a typical pre-step before using intersection functions.

    Args:
        geometry: A Rhino Brep or Mesh.
        high_accuracy: If True, a physically accurate bounding box will be computed.
            If not, a bounding box estimate will be computed. For some geometry
            types, there is no difference between the estimate and the accurate
            bounding box. Estimated bounding boxes can be computed much (much)
            faster than accurate (or "tight") bounding boxes. Estimated bounding
            boxes are always similar to or larger than accurate bounding boxes.
    """
    return geometry.GetBoundingBox(high_accuracy)


def bounding_box_extents(geometry, high_accuracy=False):
    """Get min and max points around an input Rhino Mesh or Brep

    Args:
        geometry: A Rhino Brep or Mesh.
        high_accuracy: If True, a physically accurate bounding box will be computed.
            If not, a bounding box estimate will be computed. For some geometry
            types, there is no difference between the estimate and the accurate
            bounding box. Estimated bounding boxes can be computed much (much)
            faster than accurate (or "tight") bounding boxes. Estimated bounding
            boxes are always similar to or larger than accurate bounding boxes.
    """
    b_box = bounding_box(geometry, high_accuracy)
    return b_box.Max, b_box.Min


def intersect_mesh_rays(
        mesh, points, vectors, normals=None, cpu_count=None, parallel=True):
    """Intersect a group of rays (represented by points and vectors) with a mesh.

    All combinations of rays that are possible between the input points and
    vectors will be intersected. This method exists since most CAD plugins have
    much more efficient mesh/ray intersection functions than ladybug_geometry.
    However, the ladybug_geometry Face3D.intersect_line_ray() method provides
    a workable (albeit very inefficient) alternative to this if it is needed.

    Args:
        mesh: A Rhino mesh that can block the rays.
        points: An array of Rhino points that will be used to generate rays.
        vectors: An array of Rhino vectors that will be used to generate rays.
        normals: An optional array of Rhino vectors that align with the input
            points and denote the direction each point is facing. These will
            be used to eliminate any cases where the vector and the normal differ
            by more than 90 degrees. If None, points are assumed to have no direction.
        cpu_count: An integer for the number of CPUs to be used in the intersection
            calculation. The ladybug_rhino.grasshopper.recommended_processor_count
            function can be used to get a recommendation. If set to None, all
            available processors will be used. (Default: None).
        parallel: Optional boolean to override the cpu_count and use a single CPU
            instead of multiple processors.

    Returns:
        A tuple with two elements

        -   intersection_matrix -- A 2D matrix of 0's and 1's indicating the results
            of the intersection. Each sub-list of the matrix represents one of the
            points and has a length equal to the vectors. 0 indicates a blocked
            ray and 1 indicates a ray that was not blocked.

        -   angle_matrix -- A 2D matrix of angles in radians. Each sub-list of the
            matrix represents one of the normals and has a length equal to the
            supplied vectors. Will be None if no normals are provided.
    """
    intersection_matrix = [0] * len(points)  # matrix to be filled with results
    angle_matrix = [0] * len(normals) if normals is not None else None
    cutoff_angle = math.pi / 2  # constant used in all normal checks
    if not parallel:
        cpu_count = 1

    def intersect_point(i):
        """Intersect all of the vectors of a given point without any normal check."""
        pt = points[i]
        int_list = []
        for vec in vectors:
            ray = rg.Ray3d(pt, vec)
            if rg.Intersect.Intersection.MeshRay(mesh, ray) >= 0:
                is_clear = 0
            else:
                is_clear = 1
            int_list.append(is_clear)
        intersection_matrix[i] = int_list

    def intersect_point_normal_check(i):
        """Intersect all of the vectors of a given point with a normal check."""
        pt, normal_vec = points[i], normals[i]
        int_list = []
        angle_list = []
        for vec in vectors:
            vec_angle = rg.Vector3d.VectorAngle(normal_vec, vec)
            angle_list.append(vec_angle)
            if vec_angle <= cutoff_angle:
                ray = rg.Ray3d(pt, vec)
                if rg.Intersect.Intersection.MeshRay(mesh, ray) >= 0:
                    is_clear = 0
                else:
                    is_clear = 1
                int_list.append(is_clear)
            else:  # the vector is pointing behind the surface
                int_list.append(0)
        intersection_matrix[i] = specializedarray.array('B', int_list)
        angle_matrix[i] = specializedarray.array('d', angle_list)

    def intersect_each_point_group(worker_i):
        """Intersect groups of points so that only the cpu_count is used."""
        start_i, stop_i = pt_groups[worker_i]
        for count in range(start_i, stop_i):
            intersect_point(count)

    def intersect_each_point_group_normal_check(worker_i):
        """Intersect groups of points with distance check so only cpu_count is used."""
        start_i, stop_i = pt_groups[worker_i]
        for count in range(start_i, stop_i):
            intersect_point_normal_check(count)

    if cpu_count is not None and cpu_count > 1:
        # group the points in order to meet the cpu_count
        pt_count = len(points)
        worker_count = min((cpu_count, pt_count))
        i_per_group = int(math.ceil(pt_count / worker_count))
        pt_groups = [[x, x + i_per_group] for x in range(0, pt_count, i_per_group)]
        pt_groups[-1][-1] = pt_count  # ensure the last group ends with point count

    if normals is not None:
        if cpu_count is None:  # use all available CPUs
            tasks.Parallel.ForEach(range(len(points)), intersect_point_normal_check)
        elif cpu_count <= 1:  # run everything on a single processor
            for i in range(len(points)):
                intersect_point_normal_check(i)
        else:  # run the groups in a manner that meets the CPU count
            tasks.Parallel.ForEach(
                range(len(pt_groups)), intersect_each_point_group_normal_check)
    else:
        if cpu_count is None:  # use all available CPUs
            tasks.Parallel.ForEach(range(len(points)), intersect_point)
        elif cpu_count <= 1:  # run everything on a single processor
            for i in range(len(points)):
                intersect_point(i)
        else:  # run the groups in a manner that meets the CPU count
            tasks.Parallel.ForEach(range(len(pt_groups)), intersect_each_point_group)

    return intersection_matrix, angle_matrix


def intersect_mesh_lines(
        mesh, start_points, end_points, max_dist=None, cpu_count=None, parallel=True):
    """Intersect a group of lines (represented by start + end points) with a mesh.

    All combinations of lines that are possible between the input start_points and
    end_points will be intersected. This method exists since most CAD plugins have
    much more efficient mesh/line intersection functions than ladybug_geometry.
    However, the ladybug_geometry Face3D.intersect_line_ray() method provides
    a workable (albeit very inefficient) alternative to this if it is needed.

    Args:
        mesh: A Rhino mesh that can block the lines.
        start_points: An array of Rhino points that will be used to generate lines.
        end_points: An array of Rhino points that will be used to generate lines.
        max_dist: An optional number to set the maximum distance beyond which the
            end_points are no longer considered visible by the start_points.
            If None, points with an unobstructed view to one another will be
            considered visible no matter how far they are from one another.
        cpu_count: An integer for the number of CPUs to be used in the intersection
            calculation. The ladybug_rhino.grasshopper.recommended_processor_count
            function can be used to get a recommendation. If set to None, all
            available processors will be used. (Default: None).
        parallel: Optional boolean to override the cpu_count and use a single CPU
            instead of multiple processors.

    Returns:
        A 2D matrix of 0's and 1's indicating the results of the intersection.
        Each sub-list of the matrix represents one of the points and has a
        length equal to the end_points. 0 indicates a blocked ray and 1 indicates
        a ray that was not blocked.
    """
    int_matrix = [0] * len(start_points)  # matrix to be filled with results
    if not parallel:
        cpu_count = 1

    def intersect_line(i):
        """Intersect a line defined by a start and an end with the mesh."""
        pt = start_points[i]
        int_list = []
        for ept in end_points:
            lin = rg.Line(pt, ept)
            int_obj = rg.Intersect.Intersection.MeshLine(mesh, lin)
            is_clear = 1 if None in int_obj or len(int_obj) == 0 else 0
            int_list.append(is_clear)
        int_matrix[i] = int_list

    def intersect_line_dist_check(i):
        """Intersect a line with the mesh with a distance check."""
        pt = start_points[i]
        int_list = []
        for ept in end_points:
            lin = rg.Line(pt, ept)
            if lin.Length > max_dist:
                int_list.append(0)
            else:
                int_obj = rg.Intersect.Intersection.MeshLine(mesh, lin)
                is_clear = 1 if None in int_obj or len(int_obj) == 0 else 0
                int_list.append(is_clear)
        int_matrix[i] = int_list

    def intersect_each_line_group(worker_i):
        """Intersect groups of lines so that only the cpu_count is used."""
        start_i, stop_i = l_groups[worker_i]
        for count in range(start_i, stop_i):
            intersect_line(count)

    def intersect_each_line_group_dist_check(worker_i):
        """Intersect groups of lines with distance check so only cpu_count is used."""
        start_i, stop_i = l_groups[worker_i]
        for count in range(start_i, stop_i):
            intersect_line_dist_check(count)

    if cpu_count is not None and cpu_count > 1:
        # group the lines in order to meet the cpu_count
        l_count = len(start_points)
        worker_count = min((cpu_count, l_count))
        i_per_group = int(math.ceil(l_count / worker_count))
        l_groups = [[x, x + i_per_group] for x in range(0, l_count, i_per_group)]
        l_groups[-1][-1] = l_count  # ensure the last group ends with line count

    if max_dist is not None:
        if cpu_count is None:  # use all available CPUs
            tasks.Parallel.ForEach(range(len(start_points)), intersect_line_dist_check)
        elif cpu_count <= 1:  # run everything on a single processor
            for i in range(len(start_points)):
                intersect_line_dist_check(i)
        else:  # run the groups in a manner that meets the CPU count
            tasks.Parallel.ForEach(
                range(len(l_groups)), intersect_each_line_group_dist_check)
    else:
        if cpu_count is None:  # use all available CPUs
            tasks.Parallel.ForEach(range(len(start_points)), intersect_line)
        elif cpu_count <= 1:  # run everything on a single processor
            for i in range(len(start_points)):
                intersect_line(i)
        else:  # run the groups in a manner that meets the CPU count
            tasks.Parallel.ForEach(
                range(len(l_groups)), intersect_each_line_group)
    return int_matrix


def trace_ray(ray, breps, bounce_count=1):
    """Get a list of Rhino points for the path a ray takes bouncing through breps.

    Args:
        ray: A Rhino Ray whose path will be traced through the geometry.
        breps: An array of Rhino breps through with the ray will be traced.
        bounce_count: An positive integer for the number of ray bounces to trace
            the sun rays forward. (Default: 1).
    """
    return rg.Intersect.Intersection.RayShoot(ray, breps, bounce_count)


def normal_at_point(brep, point):
    """Get a Rhino vector for the normal at a specific point that lies on a brep.

    Args:
        breps: A Rhino brep on which the normal direction will be evaluated.
        point: A Rhino point on the input brep where the normal will be evaluated.
    """
    return brep.ClosestPoint(point, tolerance)[5]


def intersect_solids_parallel(solids, bound_boxes, cpu_count=None):
    """Intersect the co-planar faces of an array of solids using parallel processing.

    Args:
        original_solids: An array of closed Rhino breps (polysurfaces) that do
            not have perfectly matching surfaces between adjacent Faces.
        bound_boxes: An array of Rhino bounding boxes that parallels the input
            solids and will be used to check whether two Breps have any potential
            for intersection before the actual intersection is performed.
        cpu_count: An integer for the number of CPUs to be used in the intersection
            calculation. The ladybug_rhino.grasshopper.recommended_processor_count
            function can be used to get a recommendation. If None, all available
            processors will be used. (Default: None).
        parallel: Optional boolean to override the cpu_count and use a single CPU
            instead of multiple processors.

    Returns:
        int_solids -- The input array of solids, which have all been intersected
        with one another.
    """
    int_solids = solids[:]  # copy the input list to avoid editing it

    def intersect_each_solid(i):
        """Intersect a solid with all of the other solids of the list."""
        bb_1 = bound_boxes[i]
        # intersect the solids that come after this one
        for j, bb_2 in enumerate(bound_boxes[i + 1:]):
            if not overlapping_bounding_boxes(bb_1, bb_2):
                continue  # no overlap in bounding box; intersection impossible
            split_brep1, int_exists = \
                intersect_solid(int_solids[i], int_solids[i + j + 1])
            if int_exists:
                int_solids[i] = split_brep1
        # intersect the solids that come before this one
        for j, bb_2 in enumerate(bound_boxes[:i]):
            if not overlapping_bounding_boxes(bb_1, bb_2):
                continue  # no overlap in bounding box; intersection impossible
            split_brep2, int_exists = intersect_solid(int_solids[i], int_solids[j])
            if int_exists:
                int_solids[i] = split_brep2

    def intersect_each_solid_group(worker_i):
        """Intersect groups of solids so that only the cpu_count is used."""
        start_i, stop_i = s_groups[worker_i]
        for count in range(start_i, stop_i):
            intersect_each_solid(count)

    if cpu_count is None or cpu_count <= 1:  # use all available CPUs
        tasks.Parallel.ForEach(range(len(solids)), intersect_each_solid)
    else:  # group the solids in order to meet the cpu_count
        solid_count = len(int_solids)
        worker_count = min((cpu_count, solid_count))
        i_per_group = int(math.ceil(solid_count / worker_count))
        s_groups = [[x, x + i_per_group] for x in range(0, solid_count, i_per_group)]
        s_groups[-1][-1] = solid_count  # ensure the last group ends with solid count
        tasks.Parallel.ForEach(range(len(s_groups)), intersect_each_solid_group)

    return int_solids


def intersect_solids(solids, bound_boxes):
    """Intersect the co-planar faces of an array of solids.

    Args:
        original_solids: An array of closed Rhino breps (polysurfaces) that do
            not have perfectly matching surfaces between adjacent Faces.
        bound_boxes: An array of Rhino bounding boxes that parallels the input
            solids and will be used to check whether two Breps have any potential
            for intersection before the actual intersection is performed.

    Returns:
        int_solids -- The input array of solids, which have all been intersected
        with one another.
    """
    int_solids = solids[:]  # copy the input list to avoid editing it

    for i, bb_1 in enumerate(bound_boxes):
        for j, bb_2 in enumerate(bound_boxes[i + 1:]):
            if not overlapping_bounding_boxes(bb_1, bb_2):
                continue  # no overlap in bounding box; intersection impossible

            # split the first solid with the second one
            split_brep1, int_exists = intersect_solid(
                int_solids[i], int_solids[i + j + 1])
            int_solids[i] = split_brep1

            # split the second solid with the first one if an intersection was found
            if int_exists:
                split_brep2, int_exists = intersect_solid(
                    int_solids[i + j + 1], int_solids[i])
                int_solids[i + j + 1] = split_brep2

    return int_solids


def intersect_solid(solid, other_solid):
    """Intersect the co-planar faces of one solid Brep using another.

    Args:
        solid: The solid Brep which will be split with intersections.
        other_solid: The other Brep, which will be used to split.

    Returns:
        A tuple with two elements

        -   solid -- The input solid, which has been split.

        -   intersection_exists -- Boolean to note whether an intersection was found
            between the solid and the other_solid. If False, there's no need to
            split the other_solid with the input solid.
    """
    # variables to track the splitting process
    intersection_exists = False  # boolean to note whether an intersection exists
    temp_brep = solid.Split(other_solid, tolerance)
    if len(temp_brep) != 0:
        solid = rg.Brep.JoinBreps(temp_brep, tolerance)[0]
        solid.Faces.ShrinkFaces()
        intersection_exists = True
    return solid, intersection_exists


def overlapping_bounding_boxes(bound_box1, bound_box2):
    """Check if two Rhino bounding boxes overlap within the tolerance.

    This is particularly useful as a check before performing computationally
    intense intersection processes between two bounding boxes. Checking the
    overlap of the bounding boxes is extremely quick given this method's use
    of the Separating Axis Theorem. This method is built into the intersect_solids
    functions in order to improve its calculation time.

    Args:
        bound_box1: The first bound_box to check.
        bound_box2: The second bound_box to check.
    """
    # Bounding box check using the Separating Axis Theorem
    bb1_width = bound_box1.Max.X - bound_box1.Min.X
    bb2_width = bound_box2.Max.X - bound_box2.Min.X
    dist_btwn_x = abs(bound_box1.Center.X - bound_box2.Center.X)
    x_gap_btwn_box = dist_btwn_x - (0.5 * bb1_width) - (0.5 * bb2_width)

    bb1_depth = bound_box1.Max.Y - bound_box1.Min.Y
    bb2_depth = bound_box2.Max.Y - bound_box2.Min.Y
    dist_btwn_y = abs(bound_box1.Center.Y - bound_box2.Center.Y)
    y_gap_btwn_box = dist_btwn_y - (0.5 * bb1_depth) - (0.5 * bb2_depth)

    bb1_height = bound_box1.Max.Z - bound_box1.Min.Z
    bb2_height = bound_box2.Max.Z - bound_box2.Min.Z
    dist_btwn_z = abs(bound_box1.Center.Z - bound_box2.Center.Z)
    z_gap_btwn_box = dist_btwn_z - (0.5 * bb1_height) - (0.5 * bb2_height)

    if x_gap_btwn_box > tolerance or y_gap_btwn_box > tolerance or \
            z_gap_btwn_box > tolerance:
        return False  # no overlap
    return True  # overlap exists


def split_solid_to_floors(building_solid, floor_heights):
    """Extract a series of planar floor surfaces from solid building massing.

    Args:
        building_solid: A closed brep representing a building massing.
        floor_heights: An array of float values for the floor heights, which
            will be used to generate planes that subdivide the building solid.

    Returns:
        floor_breps -- A list of planar, horizontal breps representing the floors
        of the building.
    """
    # get the floor brep at each of the floor heights.
    floor_breps = []
    for hgt in floor_heights:
        story_breps = []
        floor_base_pt = rg.Point3d(0, 0, hgt)
        section_plane = rg.Plane(floor_base_pt, rg.Vector3d.ZAxis)
        floor_crvs = rg.Brep.CreateContourCurves(building_solid, section_plane)
        try:  # Assume a single contour curve has been found
            floor_brep = rg.Brep.CreatePlanarBreps(floor_crvs, tolerance)
        except TypeError:  # An array of contour curves has been found
            floor_brep = rg.Brep.CreatePlanarBreps(floor_crvs)
        if floor_brep is not None:
            story_breps.extend(floor_brep)
        floor_breps.append(story_breps)

    return floor_breps


def geo_min_max_height(geometry):
    """Get the min and max Z values of any input object.

    This is useful as a pre-step before the split_solid_to_floors method.
    """
    # intersection functions changed in Rhino 7.15 such that we now need 2* tolerance
    add_val = tolerance * 2 if (7, 15) <= rhino_version < (7, 17) else 0
    bound_box = geometry.GetBoundingBox(rg.Plane.WorldXY)
    return bound_box.Min.Z + add_val, bound_box.Max.Z



"""Functions to translate entire Ladybug core objects to Rhino geometries.

The methods here are intended to help translate groups of geometry that are commonly
generated by several objects in Ladybug core (ie. legends, compasses, etc.)
"""
import math

from .fromgeometry import from_mesh3d, from_arc2d, from_linesegment2d
from .text import text_objects

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError("Failed to import ladybug_geometry.\n{}".format(e))


def legend_objects(legend):
    """Translate a Ladybug Legend object into Grasshopper geometry.

    Args:
        legend: A Ladybug Legend object to be converted to Rhino geometry.

    Returns:
        A list of Rhino geometries in the following order.

        -   legend_mesh -- A colored mesh for the legend.

        -   legend_title -- A bake-able text object for the legend title.

        -   legend_text -- Bake-able text objects for the rest of the legend text.
    """
    _height = legend.legend_parameters.text_height
    _font = legend.legend_parameters.font
    legend_mesh = from_mesh3d(legend.segment_mesh)
    legend_title = text_objects(legend.title, legend.title_location, _height, _font)
    if legend.legend_parameters.continuous_legend is False:
        legend_text = [text_objects(txt, loc, _height, _font, 0, 5) for txt, loc in
                       zip(legend.segment_text, legend.segment_text_location)]
    elif legend.legend_parameters.vertical is True:
        legend_text = [text_objects(txt, loc, _height, _font, 0, 3) for txt, loc in
                       zip(legend.segment_text, legend.segment_text_location)]
    else:
        legend_text = [text_objects(txt, loc, _height, _font, 1, 5) for txt, loc in
                       zip(legend.segment_text, legend.segment_text_location)]
    return [legend_mesh] + [legend_title] + legend_text


def compass_objects(compass, z=0, custom_angles=None, projection=None, font='Arial'):
    """Translate a Ladybug Compass object into Grasshopper geometry.

    Args:
        compass: A Ladybug Compass object to be converted to Rhino geometry.
        z: A number for the Z-coordinate to be used in translation. (Default: 0)
        custom_angles: An array of numbers between 0 and 360 to be used to
            generate custom angle labels around the compass.
        projection: Text for the name of the projection to use from the sky
            dome hemisphere to the 2D plane. If None, no altitude circles o
            labels will be drawn (Default: None). Choose from the following:

            * Orthographic
            * Stereographic

        font: Optional text for the font to be used in creating the text.
            (Default: 'Arial')

    Returns:
        A list of Rhino geometries in the following order.

        -   all_boundary_circles -- Three Circle objects for the compass boundary.

        -   major_azimuth_ticks -- Line objects for the major azimuth labels.

        -   major_azimuth_text -- Bake-able text objects for the major azimuth labels.

        -   minor_azimuth_ticks -- Line objects for the minor azimuth labels
                (if applicable).

        -   minor_azimuth_text -- Bake-able text objects for the minor azimuth
                labels (if applicable).

        -   altitude_circles -- Circle objects for the altitude labels.

        -   altitude_text -- Bake-able text objects for the altitude labels.

     """
    # set default variables based on the compass properties
    maj_txt = compass.radius / 20
    min_txt = maj_txt / 2
    xaxis = Vector3D(1, 0, 0).rotate_xy(math.radians(compass.north_angle))

    result = []  # list to hold all of the returned objects
    for circle in compass.all_boundary_circles:
        result.append(from_arc2d(circle, z))

    # generate the labels and tick marks for the azimuths
    if custom_angles is None:
        for line in compass.major_azimuth_ticks:
            result.append(from_linesegment2d(line, z))
        for txt, pt in zip(compass.MAJOR_TEXT, compass.major_azimuth_points):
            txt_pln = Plane(o=Point3D(pt.x, pt.y, z), x=xaxis)
            result.append(text_objects(txt, txt_pln, maj_txt, font, 1, 3))
        for line in compass.minor_azimuth_ticks:
            result.append(from_linesegment2d(line, z))
        for txt, pt in zip(compass.MINOR_TEXT, compass.minor_azimuth_points):
            txt_pln = Plane(o=Point3D(pt.x, pt.y, z), x=xaxis)
            result.append(text_objects(txt, txt_pln, min_txt, font, 1, 3))
    else:
        for line in compass.ticks_from_angles(custom_angles):
            result.append(from_linesegment2d(line, z))
        for txt, pt in zip(
                custom_angles, compass.label_points_from_angles(custom_angles)):
            txt_pln = Plane(o=Point3D(pt.x, pt.y, z), x=xaxis)
            result.append(text_objects(str(txt), txt_pln, maj_txt, font, 1, 3))

    # generate the labels and tick marks for the altitudes
    if projection is not None:
        if projection.title() == 'Orthographic':
            for circle in compass.orthographic_altitude_circles:
                result.append(from_arc2d(circle, z))
            for txt, pt in zip(compass.ALTITUDES, compass.orthographic_altitude_points):
                txt_pln = Plane(o=Point3D(pt.x, pt.y, z), x=xaxis)
                result.append(text_objects(str(txt), txt_pln, min_txt, font, 1, 0))
        elif projection.title() == 'Stereographic':
            for circle in compass.stereographic_altitude_circles:
                result.append(from_arc2d(circle, z))
            for txt, pt in zip(compass.ALTITUDES, compass.stereographic_altitude_points):
                txt_pln = Plane(o=Point3D(pt.x, pt.y, z), x=xaxis)
                result.append(text_objects(str(txt), txt_pln, min_txt, font, 1, 0))

    return result



"""Ladybug_rhino configurations.

Global variables such as tolerances, units and Rhino versions are stored here.
"""
import os

try:
    from ladybug.config import folders as lb_folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    import Rhino
    rhino_version_str = str(Rhino.RhinoApp.Version)
    rhino_version = tuple(int(n) for n in rhino_version_str.split('.'))
except Exception:  # Rhino is unavailable; just use a placeholder
    rhino_version = (7, 0)

try:  # Try to import tolerance from the active Rhino document
    import scriptcontext
    tolerance = scriptcontext.doc.ModelAbsoluteTolerance
    angle_tolerance = scriptcontext.doc.ModelAngleToleranceDegrees
except ImportError:  # No Rhino doc is available. Use Rhino's default.
    tolerance = 0.01
    angle_tolerance = 1.0  # default is 1 degree

from .ghpath import find_grasshopper_userobjects, find_grasshopper_libraries


def conversion_to_meters():
    """Get the conversion factor to meters based on the current Rhino doc units system.

    Returns:
        A number for the conversion factor, which should be multiplied by all distance
        units taken from Rhino geometry in order to convert them to meters.
    """
    try:  # Try to import units from the active Rhino document
        import scriptcontext
        units = str(scriptcontext.doc.ModelUnitSystem).split('.')[-1]
    except ImportError:  # No Rhino doc available. Default to the greatest of all units
        units = 'Meters'

    if units == 'Meters':
        return 1.0
    elif units == 'Millimeters':
        return 0.001
    elif units == 'Feet':
        return 0.305
    elif units == 'Inches':
        return 0.0254
    elif units == 'Centimeters':
        return 0.01
    else:
        raise ValueError(
            "You're kidding me! What units are you using?" + units + "?\n"
            "Please use Meters, Millimeters, Centimeters, Feet or Inches.")


def units_system():
    """Get text for the current Rhino doc units system. (eg. 'Meters', 'Feet')"""
    try:  # Try to import units from the active Rhino document
        import scriptcontext
        return str(scriptcontext.doc.ModelUnitSystem).split('.')[-1]
    except ImportError:  # No Rhino doc available. Default to the greatest of all units
        return 'Meters'


def units_abbreviation():
    """Get text for the current Rhino doc units abbreviation (eg. 'm', 'ft')"""
    try:  # Try to import units from the active Rhino document
        import scriptcontext
        units = str(scriptcontext.doc.ModelUnitSystem).split('.')[-1]
    except ImportError:  # No Rhino doc available. Default to the greatest of all units
        units = 'Meters'

    if units == 'Meters':
        return 'm'
    elif units == 'Millimeters':
        return 'mm'
    elif units == 'Feet':
        return 'ft'
    elif units == 'Inches':
        return 'in'
    elif units == 'Centimeters':
        return 'cm'
    else:
        raise ValueError(
            "You're kidding me! What units are you using?" + units + "?\n"
            "Please use Meters, Millimeters, Centimeters, Feet or Inches.")


class Folders(object):
    """Ladybug-rhino folders.

    Properties:
        * uo_folder
        * gha_folder
        * lbt_grasshopper_version
        * lbt_grasshopper_version_str
    """

    def __init__(self):
        # find the location where the Grasshopper user objects are stored
        self._uo_folder = find_grasshopper_userobjects()[-1]
        self._gha_folder = find_grasshopper_libraries()[-1]
        if os.name == 'nt':
            # test to see if components live in the core installation
            lbt_components = os.path.join(lb_folders.ladybug_tools_folder, 'grasshopper')
            if os.path.isdir(lbt_components):
                user_dir = os.path.join(self._uo_folder, 'ladybug_grasshopper')
                if not os.path.isdir(user_dir):
                    self._uo_folder = lbt_components
                    self._gha_folder = lbt_components
        self._lbt_grasshopper_version = None
        self._lbt_grasshopper_version_str = None

    @property
    def uo_folder(self):
        """Get the path to the user object folder."""
        return self._uo_folder

    @property
    def gha_folder(self):
        """Get the path to the GHA Grasshopper component folder."""
        return self._gha_folder

    @property
    def lbt_grasshopper_version(self):
        """Get a tuple for the version of lbt-grasshopper (eg. (3, 8, 2)).

        This will be None if the version could not be sensed.
        """
        if self._lbt_grasshopper_version is None:
            self._lbt_grasshopper_version_from_txt()
        return self._lbt_grasshopper_version

    @property
    def lbt_grasshopper_version_str(self):
        """Get text for the full version of python (eg."3.8.2").

        This will be None if the version could not be sensed.
        """
        if self._lbt_grasshopper_version_str is None:
            self._lbt_grasshopper_version_from_txt()
        return self._lbt_grasshopper_version_str

    def _lbt_grasshopper_version_from_txt(self):
        """Get the LBT-Grasshopper version from the requirements.txt file in uo_folder.
        """
        req_file = os.path.join(self._uo_folder, 'requirements.txt')
        if os.path.isfile(req_file):
            with open(req_file) as rf:
                for row in rf:
                    if row.startswith('lbt-grasshopper=='):
                        lbt_ver = row.split('==')[-1].strip()
                        try:
                            self._lbt_grasshopper_version = \
                                tuple(int(i) for i in lbt_ver.split('.'))
                            self._lbt_grasshopper_version_str = lbt_ver
                        except Exception:
                            pass  # failed to parse the version into values
                        break


"""Object possessing all key folders within the configuration."""
folders = Folders()



"""Collection of methods for converting between Ladybug and .NET colors."""

try:
    from System.Drawing import Color
except ImportError as e:
    raise ImportError("Failed to import Windows/.NET libraries\n{}".format(e))


def color_to_color(color, alpha=255):
    """Convert a ladybug color into .NET color.

    Args:
        alpha: Optional integer between 1 and 255 for the alpha value of the color.
    """
    try:
        return Color.FromArgb(alpha, color.r, color.g, color.b)
    except AttributeError as e:
        raise AttributeError('Input must be of type of Color:\n{}'.format(e))


def argb_color_to_color(color):
    """Convert a ladybug color into .NET color, including the alpha channel."""
    try:
        return Color.FromArgb(color.a, color.r, color.g, color.b)
    except AttributeError as e:
        raise AttributeError('Input must be of type of Color:\n{}'.format(e))


def gray():
    """Get a .NET gray color object. Useful when you need a placeholder color."""
    return Color.Gray


def black():
    """Get a .NET black color object. Useful for things like default text."""
    return Color.Black



"""Functions for setting lights within the Rhino scene."""
from ladybug.dt import DateTime

try:
    import System
except ImportError as e:  # No .NET; We are really screwed
    raise ImportError("Failed to import System.\n{}".format(e))

try:
    import Rhino.Render.Sun as sun
    import Rhino.RhinoDoc as rhdoc
    doc = rhdoc.ActiveDoc
except ImportError as e:
    raise ImportError("Failed to import Rhino document attributes.\n{}".format(e))


def set_sun(location, hoy, north=0):
    """Set the sun in the Rhino scene to correspond to a given location and DateTime.

    Args:
        location: A Ladybug Location object to set the latitude, longitude and
            time zone of the Rhino sun path.
        hoy: A number between 0 and 8760 that represent the hour of the year at
            which to evaluate the sun position. Note that this does not need to
            be an integer and decimal values can be used to specify date times
            that are not on the hour mark.
        north: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. (Default: 0).

    Returns:
        The Rhino sun object.
    """
    # process the hoy into a .NET date/time
    lb_dt = DateTime.from_hoy(hoy)
    rh_dt = System.DateTime(
        lb_dt.year, lb_dt.month, lb_dt.day, lb_dt.hour, lb_dt.minute, 0)

    # enable the sun and set its position based on the location and date/time
    sun_position = doc.Lights.Sun
    sun.Enabled.SetValue(sun_position, True)
    sun.TimeZone.SetValue(sun_position, location.time_zone)
    sun.SetPosition(sun_position, rh_dt, location.latitude, location.longitude)

    # set the north of the sun, ensuring the the y-axis is North
    sun.North.SetValue(sun_position, 90 + north)
    return sun


def disable_sun():
    """Disable the sun in the Rhino scene so it does not interfere with other lights."""
    doc.Lights.Sun.Enabled = False



"""Functions for exporting all content from Grasshopper component objects."""
import os
import shutil
import re

try:
    import System.Drawing
except ImportError:
    raise ImportError("Failed to import System.")

try:
    import Grasshopper
except ImportError:
    raise ImportError("Failed to import Grasshopper.")

from .diff import current_userobject_version, validate_version
from .userobject import create_userobject
from .component import Component

# characters that get removed and replaced when generating clean file names
REMOVE_CHARACTERS = ('LB ', 'HB ', 'DF ')
REPLACE_CHARACTERS = (' ', '/', '?', '|')

# map from the exposure category of a component to the order in which it is displayed
REVERSE_EXPOSURE_MAP = {
    'obscure': 100,
    'hidden': 100,
    'primary': 1,
    'secondary': 2,
    'tertiary': 3,
    'quarternary': 4,
    'quinary': 5,
    'senary': 6,
    'septenary': 7
}


def export_component(folder, component, change_type='fix'):
    """Export a Grasshopper component object to a package folder.

    This method writes the following files:

    * A .ghuser into the user_objects subfolder
    * A .py file into the src subfolder
    * A .json into the json subfolder
    * A .png into the icon subfolder

    Args:
        folder: Path to a folder into which the component files will be exported.
            Typically, this is the package folder of a grasshopper plugin repo.
            (eg. ladybug-grasshopper/ladybug_grasshopper).
        component: The Grasshopper component object to be exported to the folder.
        change_type: Text for the change type of the export. Valid change types
            can be seen under the CHANGE_TAGS property of the userobject module.
    """
    # process the component into a user object
    current_version = current_userobject_version(component)
    validate_version(current_version, component.Message, change_type)
    uo = create_userobject(component)

    # create subfolders in the folder if they are not already created
    uo_folder = os.path.join(folder, 'user_objects')
    src_folder = os.path.join(folder, 'src')
    json_folder = os.path.join(folder, 'json')
    icon_folder = os.path.join(folder, 'icon')
    for f in (folder, uo_folder, src_folder, json_folder, icon_folder):
        if not os.path.isdir(f):
            os.mkdir(f)

    # get the paths to the where the files will be written
    uo_fp = os.path.join(uo_folder, '%s.ghuser' % uo.Description.Name)
    src_fp = os.path.join(src_folder, '%s.py' % uo.Description.Name)
    json_fp = os.path.join(json_folder, '%s.json' % uo.Description.Name)
    icon_fp = os.path.join(icon_folder, '%s.png' % uo.Description.Name)

    # export the userobject to the user_objects subfolder
    if os.path.isfile(uo_fp):
        os.remove(uo_fp)
    uo.Path = uo_fp
    uo.SaveToFile()

    # export the .py file to the src subfolder
    code = uo.InstantiateObject().Code
    try:
        if isinstance(code, unicode):
            code = code.encode('utf8', 'ignore').replace("\r", "")
    except Exception:
        pass  # we are not in Python 2
    with open(src_fp, 'w') as outf:
        outf.write(code)

    # export the .json file to the json subfolder
    if os.path.isfile(json_fp):
        os.remove(json_fp)
    component_obj = Component.from_gh_component(component)
    component_obj.to_json(json_folder)

    # export the icon file
    icon = component.Icon_24x24
    icon.Save(icon_fp)

    print('    UserObject, source code, icon and JSON are copied to folder.')


def export_component_screen_capture(folder, component, x_dim=1000, y_dim=1000):
    """Export a screen capture of a Grasshopper component object to a folder.

    The image will always be centered on the component and at a resolution where
    the inputs and outputs are clearly visible.

    Args:
        folder: Path to a folder into which the image file will be exported.
        component: The Grasshopper component object to be exported to the folder.
        x_dim: Integer for the X dimension of the exported image in
            pixels. (Default: 1000).
        y_dim: Integer for the X dimension of the exported image in
            pixels. (Default: 1000).
    """
    # Get the coordinates of the upper-left corner of the image from the component
    if component.Name == 'LB ImageViewer':
        ul_x = component.Attributes.Pivot.X - int(((x_dim / 2) - 400) / 2)
        ul_y = component.Attributes.Pivot.Y - int(((y_dim / 2) - 400) / 2)
    else:
        ul_x = component.Attributes.Pivot.X - int(((x_dim / 2) - 120) / 2)
        ul_y = component.Attributes.Pivot.Y - int(((y_dim / 2) - 120) / 2)
    rect = System.Drawing.Rectangle(ul_x, ul_y, x_dim, y_dim)

    # set the image zoon/resolution
    image_settings = Grasshopper.GUI.Canvas.GH_Canvas.GH_ImageSettings()
    image_settings.Zoom = 1.95
    canvas = Grasshopper.GH_InstanceServer.ActiveCanvas

    # capture the image of the component
    images_of_canvas = canvas.GenerateHiResImage(rect, image_settings)
    screen_capture = images_of_canvas[0][0]

    # resize the image
    loaded_img = System.Drawing.Bitmap(screen_capture)
    new_img = loaded_img.Clone(System.Drawing.Rectangle(0, 0, x_dim, y_dim),
                               loaded_img.PixelFormat)

    # write the image to a file
    file_name = clean_component_filename(component)
    file_path = os.path.join(folder, '{}.png'.format(file_name))
    new_img.Save(file_path)

    # delete original image
    loaded_img.Dispose()
    path = os.path.split(screen_capture)[0]
    shutil.rmtree(path)
    return file_path


def export_component_icon(folder, component):
    """Export a Grasshopper component icon to a folder.

    Args:
        folder: Path to a folder into which the icon image file will be exported.
        component: The Grasshopper component object to be exported to the folder.
    """
    file_name = clean_component_filename(component)
    file_path = os.path.join(folder, '{}.png'.format(file_name))
    icon = component.Icon_24x24
    icon.Save(file_path)


def export_component_to_markdown(folder, component, github_repo=None):
    """Export a Grasshopper component's description and metadata to a Markdown file.

    Args:
        folder: Path to a folder into which the MArkdown file will be exported.
        component: The Grasshopper component object to be exported to the folder.
        github_repo: Optional URL to a GitHub repo that can be used to link the
            Markdown page to a GitHub repository.
    """
    # get the relevant name information from the component
    b_name = component.Name
    for item in REMOVE_CHARACTERS:
        b_name = b_name.replace(item, '')
    name = clean_component_filename(component)
    lines = []

    # write the lines for the header with the image, icon and source code
    lines.append('## %s\n' % b_name)
    img_text = '![](../../images/components/%s.png)\n' % name
    lines.append(img_text)
    if github_repo:
        source = '![](../../images/icons/%s.png) - [[source code]](%s/%s.py)\n' % (
            name, github_repo, component.Name.replace(' ', '%20'))
        lines.append(source)

    # write the lines for the description
    full_desc = []
    for d_l in component.Description.split('\n'):
        if ('-' in d_l or '_' in d_l or '.' in d_l) and len(d_l) <= 2:
            full_desc.append('\n\n')
        else:
            full_desc.append('{} '.format(d_l.replace('\r', '')))
    lines.append('\n{}'.format(''.join(full_desc)))

    # check to see if there are any inputs and outputs to export
    inputs_outputs_available = True
    try:
        component.Params
    except Exception:  # no inputs and outputs available
        inputs_outputs_available = False

    if inputs_outputs_available:
        # export the inputs
        lines.append('\n#### Inputs')
        for i in range(component.Params.Input.Count):
            i_name = component.Params.Input[i].NickName
            alph = ''.join(re.findall('[a-zA-Z]+', i_name))
            if len(alph) == 0:
                continue

            t = ''
            if i_name.startswith('_') and i_name.endswith('_'):
                i_name = i_name[1:-1]
            elif i_name.startswith('_'):
                i_name = i_name[1:]
                t = '[Required]'
            elif i_name.endswith('_'):
                i_name = i_name[:-1]

            full_desc = []
            for d_l in component.Params.Input[i].Description.split('\n'):
                if ('-' in d_l or '_' in d_l or '.' in d_l) and len(d_l) <= 2:
                    full_desc.append('\n')
                elif d_l.startswith('*') or d_l.startswith('-'):
                    full_desc.append('\n\n    {}'.format(d_l.replace('\r', '')))
                else:
                    full_desc.append('{} '.format(d_l.replace('\r', '')))
            line = '* ##### {} {}\n{}'.format(i_name, t, ''.join(full_desc))
            lines.append(line)

        # export the outputs
        lines.append('\n#### Outputs')
        for i in range(component.Params.Output.Count):
            o_name = component.Params.Output[i].NickName
            alph = ''.join(re.findall('[a-zA-Z]+', o_name))
            if len(alph) == 0:
                continue
            full_desc = []
            for d_l in component.Params.Output[i].Description.split('\n'):
                if ('-' in d_l or '_' in d_l or '.' in d_l) and len(d_l) <= 2:
                    full_desc.append('\n')
                elif d_l.startswith('*') or d_l.startswith('-'):
                    full_desc.append('\n\n    {}'.format(d_l.replace('\r', '')))
                else:
                    full_desc.append('{} '.format(d_l.replace('\r', '')))
            line = '* ##### {}\n{}'.format(o_name, ''.join(full_desc))
            lines.append(line)

    # write the .md file
    file_path = os.path.join(folder, '{}.md'.format(name))
    with open(file_path, 'w') as out_f:
        out_f.write('\n'.join(lines).encode('utf-8'))
    return file_path


def export_plugin_to_markdown(folder, plugin_name):
    """Export a Grasshopper plugin and its subcategories to Markdown files.

    Args:
        folder: Path to a folder into which the icon image file will be exported.
        plugin_name: Text for the name of a particular plugin (aka. insect) to
            place components from (eg. "Ladybug", "Honeybee", "HB-Energy").
    """
    # gather all of the components of a plugin into a dictionary with subcategories
    components = {}
    for proxy in Grasshopper.Instances.ComponentServer.ObjectProxies:
        if proxy.Obsolete:
            continue
        category = proxy.Desc.Category
        subcategory = proxy.Desc.SubCategory
        # check to see if the component is in the plugin
        if category.strip() == plugin_name:  # if so, organize it by sub category
            if subcategory not in components.keys():
                components[subcategory] = {}
            expo = REVERSE_EXPOSURE_MAP[str(proxy.Exposure).split(',')[-1].strip()]
            if expo not in components[subcategory].keys():
                components[subcategory][expo] = []
            name = clean_component_filename(proxy.Desc)
            components[subcategory][expo].append(name)
    sorted_sub_categories = sorted(components.keys())

    # create the summary file header
    lines = []
    read_md = '[%s Primer](README.md)' % plugin_name
    header = '# Summary\n\n* ' + read_md + '\n* [Components](text/categories/README.md)'
    lines.append(header)

    # loop through the subcategories and add them to the index
    for s_category in sorted_sub_categories:
        # write the subcategory into the summary
        clean_cat = ''.join(s_category.split()).replace('|', '_').replace('::', '_')
        line = '\t* [%s](text/categories/%s.md)' % (s_category, clean_cat)
        lines.append(line)

        # collect text lines for the subcategory page
        s_category_lines = []
        for num in sorted(components[s_category].keys()):
            for comp in sorted(components[s_category][num]):
                readable_name = comp.replace('_', ' ')
                line = '\t\t* [%s](text/components/%s.md)' % (readable_name, comp)
                icon = '* ![IMAGE](../../images/icons/%s.png)' % comp
                s_category_lines.append(
                    icon + line.replace('\t\t*', '').replace('text', '..'))
                lines.append(line)

        # write md file for the category
        file_path = os.path.join(folder, 'text', 'categories', '%s.md' % clean_cat)
        with open(file_path, 'w') as tab:
            tab.write('#### Component list:\n' + '\n'.join(s_category_lines))

    # write the summary file
    file_path = os.path.join(folder, 'SUMMARY.md')
    with open(file_path, 'w') as summary:
        summary.write('\n'.join(lines))
    return file_path


def clean_component_filename(component):
    """Get a clean filename derived from a component's name."""
    file_name = component.Name
    for item in REMOVE_CHARACTERS:
        file_name = file_name.replace(item, '')
    for item in REPLACE_CHARACTERS:
        file_name = file_name.replace(item, '_')
    return file_name


def refresh_toolbar():
    """Try to refresh the Grasshopper toolbar after exporting a component."""
    Grasshopper.Kernel.GH_ComponentServer.UpdateRibbonUI()



"""Classes to preview things in the Rhino Display Pipeline."""
from __future__ import division

from ladybug_geometry.geometry2d import Vector2D, Point2D, Ray2D, LineSegment2D, \
    Polyline2D, Arc2D, Polygon2D, Mesh2D
from ladybug_geometry.geometry3d import Vector3D, Point3D, Ray3D, Plane, LineSegment3D, \
    Polyline3D, Arc3D, Face3D, Mesh3D, Polyface3D, Sphere, Cone, Cylinder

from ladybug.graphic import GraphicContainer
from ladybug_display.altnumber import Default
from ladybug_display.geometry3d import DisplayText3D
from ladybug_display.visualization import AnalysisGeometry

from .config import units_system
from .color import color_to_color, argb_color_to_color, black
from .fromgeometry import from_point2d, from_vector2d, from_ray2d, \
    from_arc2d, from_polygon2d, from_polyline2d, from_mesh2d, \
    from_point3d, from_vector3d, from_ray3d, from_plane, \
    from_arc3d, from_polyline3d, from_mesh3d, from_face3d, from_polyface3d, \
    from_sphere, from_cone, from_cylinder

try:
    import System
except ImportError as e:
    raise ImportError("Failed to import Windows/.NET libraries\n{}".format(e))

try:
    import Rhino.Geometry as rg
    import Rhino.Display as rd
    import Rhino.DocObjects as ro
except ImportError as e:
    raise ImportError('Failed to import Rhino.\n{}'.format(e))

try:
    import scriptcontext as sc
except ImportError:
    pass  # previewing 2D legends will not be available


class VisualizationSetConduit(rd.DisplayConduit):
    """Class to preview VisualizationSet in the Rhino Display pipeline."""

    def __init__(self, visualization_set, render_3d_legend=False):
        """Initialize VisualizationSetConduit."""
        # set the primary properties
        self.vis_con = VisualizationSetConverter(visualization_set, render_3d_legend)

    def CalculateBoundingBox(self, calculateBoundingBoxEventArgs):
        """Overwrite the method that passes the bounding box to the display."""
        calculateBoundingBoxEventArgs.IncludeBoundingBox(self.vis_con.bbox)

    def DrawForeground(self, drawEventArgs):
        """Overwrite the method that draws the objects in the display."""
        # get the DisplayPipeline from the event arguments
        display = drawEventArgs.Display

        # for each object to be rendered, pass the drawing arguments
        for draw_args in self.vis_con.draw_2d_text:
            display.Draw2dText(*draw_args)
        for draw_args in self.vis_con.draw_sprite:
            display.DrawSprite(*draw_args)

    def PreDrawObjects(self, drawEventArgs):
        """Overwrite the method that draws the objects in the display."""
        # get the DisplayPipeline from the event arguments
        display = drawEventArgs.Display

        # for each object to be rendered, pass the drawing arguments
        for draw_args in self.vis_con.draw_mesh_false_colors:
            display.DrawMeshFalseColors(draw_args)
        for draw_args in self.vis_con.draw_mesh_shaded:
            display.DrawMeshFalseColors(draw_args[0])
        for draw_args in self.vis_con.draw_brep_shaded:
            display.DrawBrepShaded(*draw_args)

    def PostDrawObjects(self, drawEventArgs):
        """Overwrite the method that draws the objects in the display."""
        # get the DisplayPipeline from the event arguments
        display = drawEventArgs.Display

        # for each object to be rendered, pass the drawing arguments
        for draw_args in self.vis_con.draw_3d_text:
            display.Draw3dText(*draw_args)
        for draw_args in self.vis_con.draw_mesh_wires:
            display.DrawMeshWires(*draw_args)
        for draw_args in self.vis_con.draw_mesh_vertices:
            display.DrawMeshVertices(*draw_args)
        for draw_args in self.vis_con.draw_point:
            display.DrawPoint(*draw_args)
        for draw_args in self.vis_con.draw_arrow:
            display.DrawArrow(*draw_args)
        for draw_args in self.vis_con.draw_brep_wires:
            display.DrawBrepWires(*draw_args)
        for draw_args in self.vis_con.draw_line:
            display.DrawLine(*draw_args)
        for draw_args in self.vis_con.draw_patterned_line:
            display.DrawPatternedLine(*draw_args)
        for draw_args in self.vis_con.draw_patterned_polyline:
            display.DrawPatternedPolyline(*draw_args)
        for draw_args in self.vis_con.draw_curve:
            display.DrawCurve(*draw_args)
        for draw_args in self.vis_con.draw_circle:
            display.DrawCircle(*draw_args)
        for draw_args in self.vis_con.draw_arc:
            display.DrawArc(*draw_args)
        for draw_args in self.vis_con.draw_sphere:
            display.DrawSphere(*draw_args)
        for draw_args in self.vis_con.draw_cone:
            display.DrawCone(*draw_args)
        for draw_args in self.vis_con.draw_cylinder:
            display.DrawCylinder(*draw_args)


class VisualizationSetConverter(object):
    """Class to translate VisualizationSets to arguments for the Rhino display pipeline.

    Args:
        visualization_set: A Ladybug Display VisualizationSet object to be translated
            into arguments for the Rhino display pipeline.
        render_3d_legend: A Boolean to note whether the VisualizationSet should be
            rendered with 3D legends for any AnalysisGeometries it
            includes. (Default: False).
        render_2d_legend: A Boolean to note whether the VisualizationSet should be
            rendered with 2D screen-oriented legends for any AnalysisGeometries it
            includes. (Default: False).

    Properties:
        * vis_set
        * render_3d_legend
        * render_2d_legend
        * min_pt
        * max_pt
        * bbox
        * draw_sprite
        * draw_2d_text
        * draw_3d_text
        * draw_mesh_false_colors
        * draw_mesh_shaded
        * draw_mesh_wires
        * draw_mesh_vertices
        * draw_brep_shaded
        * draw_brep_wires
        * draw_point
        * draw_arrow
        * draw_line
        * draw_patterned_line
        * draw_patterned_polyline
        * draw_curve
        * draw_circle
        * draw_arc
        * draw_sphere
        * draw_cone
        * draw_cylinder
    """
    TEXT_HORIZ = {
        'Left': ro.TextHorizontalAlignment.Left,
        'Center': ro.TextHorizontalAlignment.Center,
        'Right': ro.TextHorizontalAlignment.Right
    }
    TEXT_VERT = {
        'Top': ro.TextVerticalAlignment.Top,
        'Middle': ro.TextVerticalAlignment.Middle,
        'Bottom': ro.TextVerticalAlignment.Bottom
    }
    LINE_TYPE = {
        'Dashed': int('0x1C7', base=16),
        'Dotted': int('0x00001111', base=16),
        'DashDot': int('0x1F11F1', base=16)
    }

    def __init__(
            self, visualization_set, render_3d_legend=False, render_2d_legend=False):
        """Initialize VisualizationSetConverter."""
        # set the primary properties
        self.vis_set = visualization_set
        self.render_3d_legend = bool(render_3d_legend)
        self.render_2d_legend = bool(render_2d_legend)

        # get the active viewport width and height if 2D legends are requested
        if self.render_2d_legend:
            active_view = sc.doc.Views.ActiveView.ActiveViewport
            v_size = active_view.Size
            self.view_width = v_size.Width
            self.view_height = v_size.Height
        else:  # just use some defaults as it's probably not critical
            self.view_width, self.view_height = 800, 600

        # ensure the visualization set is in Rhino model units
        units_sys = units_system()
        if self.vis_set.units is not None and self.vis_set.units != units_sys:
            self.vis_set.convert_to_units(units_sys)

        # set up the bounding box and min/max point
        self.min_pt = self.vis_set.min_point
        self.max_pt = self.vis_set.max_point
        if self.render_3d_legend:  # leave extra room for legend
            center = Point3D(
                (self.min_pt.x + self.max_pt.x) / 2,
                (self.min_pt.y + self.max_pt.y) / 2,
                (self.min_pt.z + self.max_pt.z) / 2)
            self.bbox = rg.BoundingBox(
                from_point3d(self.min_pt.scale(2, center)),
                from_point3d(self.max_pt.scale(2, center)))
        else:
            self.bbox = rg.BoundingBox(
                from_point3d(self.min_pt), from_point3d(self.max_pt))

        # translate all of the rhino objects to be rendered in the scene
        self.translate_objects()
        self.sort_shaded_objects()

    def translate_objects(self):
        """Translate all of the objects in the VisualizationSet into a Rhino equivalent.

        This is a necessary pre-step before frames of the visualization set
        can be rendered in the display pipeline.
        """
        # initialize all of the lists to hold the drawing arguments
        self.draw_sprite = []
        self.draw_2d_text = []
        self.draw_3d_text = []
        self.draw_mesh_false_colors = []
        self.draw_mesh_shaded = []
        self.draw_mesh_wires = []
        self.draw_mesh_vertices = []
        self.draw_brep_shaded = []
        self.draw_brep_wires = []
        self.draw_point = []
        self.draw_arrow = []
        self.draw_line = []
        self.draw_patterned_line = []
        self.draw_patterned_polyline = []
        self.draw_curve = []
        self.draw_circle = []
        self.draw_arc = []
        self.draw_sphere = []
        self.draw_cone = []
        self.draw_cylinder = []

        # loop through visualization geometry objects and draw them
        default_leg_x3d, default_leg_x2d, default_leg_y2d = 0, 10, 50
        for geo in self.vis_set.geometry:
            if geo.hidden:
                continue
            # translate it as AnalysisGeometry if specified
            if isinstance(geo, AnalysisGeometry):
                # generate the colors that correspond to the values
                data = geo.data_sets[geo.active_data]
                graphic = GraphicContainer(
                    data.values, self.min_pt, self.max_pt,
                    data.legend_parameters, data.data_type, data.unit)
                colors = graphic.value_colors
                # translate the analysis geometry using the matching method
                if geo.matching_method == 'faces':
                    c_count = 0
                    for mesh in geo.geometry:
                        mesh.colors = colors[c_count:c_count + len(mesh.faces)]
                        c_count += len(mesh.faces)
                        self.translate_analysis_mesh(mesh, geo.display_mode)
                elif geo.matching_method == 'vertices':
                    c_count = 0
                    for mesh in geo.geometry:
                        mesh.colors = colors[c_count:c_count + len(mesh.vertices)]
                        c_count += len(mesh.vertices)
                        self.translate_analysis_mesh(mesh, geo.display_mode)
                else:  # one color per geometry object
                    for geo_obj, col in zip(geo.geometry, colors):
                        self.translate_analysis_geometry(
                            geo_obj, col, geo.display_mode)
                # if the object is set to translate 3D legends, then display it
                if self.render_3d_legend:
                    # ensure multiple legends are not on top of each other
                    if graphic.legend_parameters.is_base_plane_default:
                        l_par = graphic.legend_parameters
                        m_vec = Vector3D(default_leg_x3d, 0, 0)
                        l_par.base_plane = l_par.base_plane.move(m_vec)
                        l_par.properties_3d._is_base_plane_default = True
                        leg_width = l_par.segment_width + 6 * l_par.text_height \
                            if l_par.vertical else \
                            l_par.segment_width * (l_par.segment_count + 2)
                        default_leg_x3d += leg_width
                    self.translate_legend3d(graphic.legend)
                # if the object is set to translate 2D legends, then display it
                if self.render_2d_legend:
                    # ensure multiple legends are not on top of each other
                    l_par = graphic.legend_parameters
                    if l_par.vertical and l_par.is_origin_x_default:
                        l_par.origin_x = '{}px'.format(default_leg_x2d)
                        l_par.properties_2d._is_origin_x_default = True
                        sw = graphic.legend.parse_dim_2d(
                            l_par.segment_width_2d, self.view_width)
                        th = graphic.legend.parse_dim_2d(
                            l_par.text_height_2d, self.view_height)
                        leg_width = sw + 6 * th
                        default_leg_x2d += leg_width
                    elif not l_par.vertical and l_par.is_origin_y_default:
                        l_par.origin_y = '{}px'.format(default_leg_y2d)
                        l_par.properties_2d._is_origin_y_default = True
                        sh = graphic.legend.parse_dim_2d(
                            l_par.segment_height_2d, self.view_height)
                        th = graphic.legend.parse_dim_2d(
                            l_par.text_height_2d, self.view_height)
                        leg_height = sh + 4 * th
                        default_leg_y2d += leg_height
                    self.translate_legend2d(graphic.legend)
            else:  # it's a ContextGeometry object
                for display_obj in geo.geometry:
                    if isinstance(display_obj, DisplayText3D):
                        self.translate_display_text3d(display_obj)
                    else:
                        self.translate_context_geometry(display_obj)

    def translate_display_text3d(self, text_obj):
        """Translate ladybug DisplayText3D into arguments for drawing in the scene.

        Args:
            text_obj: The ladybug DisplayText3D to be translated.
        """
        self.draw_3d_text.append((
            text_obj.text, color_to_color(text_obj.color), from_plane(text_obj.plane),
            text_obj.height, text_obj.font, False, False,
            self.TEXT_HORIZ[text_obj.horizontal_alignment],
            self.TEXT_VERT[text_obj.vertical_alignment]))

    def translate_legend3d(self, legend):
        """Translate a ladybug Legend into arguments for drawing in the scene.

        Args:
            legend: A Ladybug Legend object to be converted to Rhino geometry.
        """
        # translate the legend mesh
        rh_mesh = from_mesh3d(legend.segment_mesh)
        self.draw_mesh_false_colors.append(rh_mesh)
        self.draw_mesh_wires.append((rh_mesh, black()))

        # translate the legend text
        _height = legend.legend_parameters.text_height
        _font = legend.legend_parameters.font
        if legend.legend_parameters.continuous_legend is False:
            legend_text = [
                DisplayText3D(txt, loc, _height, None, _font, 'Left', 'Bottom')
                for txt, loc in zip(legend.segment_text, legend.segment_text_location)]
        elif legend.legend_parameters.vertical is True:
            legend_text = [
                DisplayText3D(txt, loc, _height, None, _font, 'Left', 'Middle')
                for txt, loc in zip(legend.segment_text, legend.segment_text_location)]
        else:
            legend_text = [
                DisplayText3D(txt, loc, _height, None, _font, 'Center', 'Bottom')
                for txt, loc in zip(legend.segment_text, legend.segment_text_location)]
        legend_title = DisplayText3D(
            legend.title, legend.title_location, _height, None, _font)
        legend_text.insert(0, legend_title)
        for txt_obj in legend_text:
            self.translate_display_text3d(txt_obj)

    def translate_legend2d(self, legend):
        """Translate a ladybug Legend into arguments for drawing in the scene.

        Args:
            legend: A Ladybug Legend object to be converted to 2D
                screen-oriented geometry.
        """
        # translate the color matrix to a bitmap
        vw, vh = self.view_width, self.view_height
        l_par = legend.legend_parameters
        color_mtx = legend.color_map_2d(vw, vh)
        color_mtx = [[color_to_color(c) for c in row] for row in color_mtx]
        net_bm = System.Drawing.Bitmap(len(color_mtx[0]), len(color_mtx))
        for y, row in enumerate(color_mtx):
            for x, col in enumerate(row):
                net_bm.SetPixel(x, y, col)
        rh_bm = rd.DisplayBitmap(net_bm)
        or_x, or_y, sh, sw, th = legend._pixel_dims_2d(vw, vh)
        s_count = l_par.segment_count
        s_count = s_count - 1 if l_par.continuous_legend else s_count
        leg_width = sw if l_par.vertical else sw * s_count
        leg_height = sh if not l_par.vertical else sh * s_count
        cent_pt = rg.Point2d(or_x + (leg_width / 2), or_y + (leg_height / 2))
        self.draw_sprite.append((rh_bm, cent_pt, leg_width, leg_height))

        # translate the legend text
        _height = legend.parse_dim_2d(l_par.text_height_2d, vh)
        _font = l_par.font
        txt_pts = legend.segment_text_location_2d(vw, vh)
        cent_txt = False if l_par.vertical else True
        legend_text = [
            (txt, black(), rg.Point2d(loc.x, loc.y), cent_txt, _height, _font)
            for txt, loc in zip(legend.segment_text, txt_pts)]
        t_pt = legend.title_location_2d(vw, vh)
        legend_title = (legend.title, black(), rg.Point2d(t_pt.x, t_pt.y),
                        False, _height, _font)
        legend_text.insert(0, legend_title)
        self.draw_2d_text.extend(legend_text)

    def translate_analysis_mesh(self, mesh, display_mode):
        """Translate an analysis mesh into arguments for drawing in the scene.

        Args:
            mesh: The Ladybug Mesh3D or Mesh2D object to be translated.
            display_mode: Text for the display mode in which to translate the mesh.
        """
        if display_mode in ('Surface', 'SurfaceWithEdges'):
            rh_mesh = from_mesh3d(mesh) if isinstance(mesh, Mesh3D) \
                else from_mesh2d(mesh)
            self.draw_mesh_false_colors.append(rh_mesh)
            if display_mode == 'SurfaceWithEdges':
                self.draw_mesh_wires.append((rh_mesh, black()))
        elif display_mode == 'Wireframe':
            rh_mesh = from_mesh3d(mesh) if isinstance(mesh, Mesh3D) \
                else from_mesh2d(mesh)
            self.draw_mesh_wires.append((rh_mesh, black()))
        elif display_mode == 'Points':
            if isinstance(mesh, Mesh3D):
                points = [from_point3d(pt) for pt in mesh.face_centroids] \
                    if mesh.is_color_by_face else \
                    [from_point3d(pt) for pt in mesh.vertices]
            else:
                points = [from_point2d(pt) for pt in mesh.face_centroids] \
                    if mesh.is_color_by_face else \
                    [from_point2d(pt) for pt in mesh.vertices]
            colors = [color_to_color(col) for col in mesh.colors]
            for col, pt in zip(colors, points):
                self.draw_point.append((pt, rd.PointStyle.RoundSimple, 3, col))

    def translate_analysis_geometry(self, geo_obj, color, display_mode):
        """Translate analysis geometry objects into arguments for drawing in the scene.

        Args:
            geo_obj: The Ladybug Geometry object to be translated.
            color: The Ladybug Color object with which the object will be translated.
            display_mode: Text for the display mode in which to translate the mesh.
        """
        # translate the color
        col = color_to_color(color)

        if isinstance(geo_obj, (Point3D, Point2D)):
            # translate analysis point
            pt = from_point3d(geo_obj) if isinstance(geo_obj, Point3D) \
                else from_point2d(geo_obj)
            self.draw_point.append((pt, rd.PointStyle.RoundSimple, 3, col))

        elif isinstance(geo_obj, (LineSegment3D, LineSegment2D)):
            # translate analysis line segment
            if isinstance(geo_obj, LineSegment3D):
                pt1, pt2 = from_point3d(geo_obj.p1), from_point3d(geo_obj.p2)
            else:
                pt1, pt2 = from_point2d(geo_obj.p1), from_point2d(geo_obj.p2)
            self.draw_line.append((pt1, pt2, col, 3))

        elif isinstance(geo_obj, (Polyline3D, Polyline2D)):
            # translate analysis polyline
            crv = from_polyline3d(geo_obj) if isinstance(geo_obj, Polyline3D) \
                else from_polyline2d(geo_obj)
            self.draw_curve.append((crv, col, 3))

        elif isinstance(geo_obj, (Face3D, Polyface3D)):
            # translate analysis Face3D and Polyface3D
            if display_mode in ('Surface', 'SurfaceWithEdges'):
                mat = rd.DisplayMaterial(col)
                rh_obj = from_face3d(geo_obj) if isinstance(geo_obj, Face3D) \
                    else from_polyface3d(geo_obj)
                self.draw_brep_shaded.append((rh_obj, mat))
                if display_mode == 'SurfaceWithEdges':
                    self.draw_brep_wires.append((rh_obj, black(), -1))
            elif display_mode == 'Wireframe':
                rh_obj = from_face3d(geo_obj) if isinstance(geo_obj, Face3D) \
                    else from_polyface3d(geo_obj)
                self.draw_brep_wires.append((rh_obj, col, 1))
            elif display_mode == 'Points':
                for pt in geo_obj.vertices:
                    self.draw_point.append(
                        (from_point3d(pt), rd.PointStyle.RoundSimple, 3, col))

        elif isinstance(geo_obj, (Mesh3D, Mesh2D)):
            # translate analysis mesh
            rh_obj = from_mesh3d(geo_obj) if isinstance(geo_obj, Mesh3D) \
                else from_mesh2d(geo_obj)
            if display_mode in ('Surface', 'SurfaceWithEdges'):
                mat = rd.DisplayMaterial(col)
                rh_obj.VertexColors.CreateMonotoneMesh(col)
                self.draw_mesh_shaded.append((rh_obj, mat))
                if display_mode == 'SurfaceWithEdges':
                    self.draw_mesh_wires.append((rh_obj, black(), 1))
            elif display_mode == 'Wireframe':
                self.draw_mesh_wires.append((rh_obj, col, 3))
            elif display_mode == 'Points':
                self.draw_mesh_vertices.append((rh_obj, col))

        elif isinstance(geo_obj, (Ray2D, Ray3D)):
            # translate analysis rays
            rh_ray = from_ray3d(geo_obj) if isinstance(geo_obj, Ray3D) \
                else from_ray2d(geo_obj)
            rh_obj = rg.Line(rh_ray.Position, rh_ray.Direction)
            self.draw_arrow.append((rh_obj, col))

        elif isinstance(geo_obj, Sphere):
            # translate analysis sphere
            if display_mode in ('Surface', 'SurfaceWithEdges'):
                rh_obj, mat = from_sphere(geo_obj), rd.DisplayMaterial(col)
                self.draw_brep_shaded.append((rh_obj.ToBrep(), mat))
                if display_mode == 'SurfaceWithEdges':
                    self.draw_sphere.append((rh_obj, black()))
            elif display_mode == 'Wireframe':
                self.draw_sphere.append((from_sphere(geo_obj), col))
            elif display_mode == 'Points':
                pt_arg = (from_point3d(geo_obj.center),
                          rd.PointStyle.RoundSimple, 3, col)
                self.draw_point.append(pt_arg)

        elif isinstance(geo_obj, (Vector2D, Vector3D)):
            # translate analysis vectors
            rh_vec = from_vector3d(geo_obj) if isinstance(geo_obj, Vector3D) \
                else from_vector2d(geo_obj)
            rh_obj = rg.Line(rg.Point3d(0, 0, 0), rh_vec)
            self.draw_arrow.append((rh_obj, col))

        elif isinstance(geo_obj, Polygon2D):
            # translate analysis polygon
            self.draw_curve.append((from_polygon2d(geo_obj), col, 3))

        elif isinstance(geo_obj, (Arc3D, Arc2D)):
            # translate analysis arcs
            crv = from_arc3d(geo_obj) if isinstance(geo_obj, Arc3D) \
                else from_arc2d(geo_obj)
            if geo_obj.is_circle:
                self.draw_circle.append((crv, col, 3))
            else:
                self.draw_arc.append((crv, col, 3))

        elif isinstance(geo_obj, Cone):
            # translate analysis cone
            if display_mode in ('Surface', 'SurfaceWithEdges'):
                rh_obj, mat = from_cone(geo_obj), rd.DisplayMaterial(col)
                self.draw_brep_shaded.append((rh_obj.ToBrep(True), mat))
                if display_mode == 'SurfaceWithEdges':
                    self.draw_cone.append((rh_obj, black()))
            elif display_mode == 'Wireframe':
                self.draw_cone.append((from_cone(geo_obj), col))
            elif display_mode == 'Points':
                pt_arg = (from_point3d(geo_obj.vertex),
                          rd.PointStyle.RoundSimple, 3, col)
                self.draw_point.append(pt_arg)

        elif isinstance(geo_obj, Cylinder):
            # translate analysis cylinder
            if display_mode in ('Surface', 'SurfaceWithEdges'):
                rh_obj, mat = from_cylinder(geo_obj), rd.DisplayMaterial(col)
                self.draw_brep_shaded.append((rh_obj.ToBrep(True, True), mat))
                if display_mode == 'SurfaceWithEdges':
                    self.draw_cylinder.append((rh_obj, black()))
            elif display_mode == 'Wireframe':
                self.draw_cylinder.append((from_cylinder(geo_obj), col))
            elif display_mode == 'Points':
                pt_arg = (from_point3d(geo_obj.center),
                          rd.PointStyle.RoundSimple, 3, col)
                self.draw_point.append(pt_arg)

        elif isinstance(geo_obj, Plane):
            # translate analysis planes
            pln = from_plane(geo_obj)
            r = 10  # default radius for a plane object in rhino model units
            # grid lines
            for j in range(-5, 6):
                i = j / 5
                p0 = pln.PointAt(i * r, -r)
                p1 = pln.PointAt(i * r, r)
                self.draw_line.append((p0, p1, col, 1))
                p0 = pln.PointAt(-r, i * r)
                p1 = pln.PointAt(r, i * r)
                self.draw_line.append((p0, p1, col, 1))
            # axes
            self.draw_line.append((pln.Origin, pln.Origin + pln.XAxis * r, col, 3))
            self.draw_line.append((pln.Origin, pln.Origin + pln.YAxis * r, col, 3))

    def translate_context_geometry(self, dis_obj):
        """Translate a display geometry object into arguments for drawing in the scene.

        Args:
            dis_obj: The Ladybug Display geometry object to be translated.
        """
        # first translate the color and get the geometry object
        col = color_to_color(dis_obj.color)
        geo_obj = dis_obj.geometry

        if isinstance(geo_obj, (Point3D, Point2D)):
            # translate display point
            pt = from_point3d(geo_obj) if isinstance(geo_obj, Point3D) \
                else from_point2d(geo_obj)
            radius = 3 if isinstance(dis_obj.radius, Default) else int(dis_obj.radius)
            self.draw_point.append((pt, rd.PointStyle.RoundSimple, radius, col))

        elif isinstance(geo_obj, (LineSegment3D, LineSegment2D)):
            # translate display line segment
            if isinstance(geo_obj, LineSegment3D):
                pt1, pt2 = from_point3d(geo_obj.p1), from_point3d(geo_obj.p2)
            else:
                pt1, pt2 = from_point2d(geo_obj.p1), from_point2d(geo_obj.p2)
            lw = 1 if isinstance(dis_obj.line_width, Default) \
                else int(dis_obj.line_width)
            if dis_obj.line_type == 'Continuous':
                self.draw_line.append((pt1, pt2, col, lw))
            else:
                self.draw_patterned_line.append(
                    (pt1, pt2, col, self.LINE_TYPE[dis_obj.line_type], lw))

        elif isinstance(geo_obj, (Polyline3D, Polyline2D)):
            # translate display polyline
            lw = 1 if isinstance(dis_obj.line_width, Default) \
                else int(dis_obj.line_width)
            if dis_obj.line_type == 'Continuous':
                crv = from_polyline3d(geo_obj) if isinstance(geo_obj, Polyline3D) \
                    else from_polyline2d(geo_obj)
                self.draw_curve.append((crv, col, lw))
            else:  # ensure the line pattern is respected
                verts = [from_point3d(pt) for pt in geo_obj.vertices] \
                    if isinstance(geo_obj, Polyline3D) else \
                    [from_point2d(pt) for pt in geo_obj.vertices]
                self.draw_patterned_polyline.append(
                    (verts, col, self.LINE_TYPE[dis_obj.line_type], lw, False))

        elif isinstance(geo_obj, (Face3D, Polyface3D)):
            # translate display Face3D and Polyface3D
            if dis_obj.display_mode in ('Surface', 'SurfaceWithEdges'):
                mat = rd.DisplayMaterial(col, 1 - (dis_obj.color.a / 255))
                rh_obj = from_face3d(geo_obj) if isinstance(geo_obj, Face3D) \
                    else from_polyface3d(geo_obj)
                self.draw_brep_shaded.append((rh_obj, mat))
                if dis_obj.display_mode == 'SurfaceWithEdges':
                    self.draw_brep_wires.append((rh_obj, black(), -1))
            elif dis_obj.display_mode == 'Wireframe':
                rh_obj = from_face3d(geo_obj) if isinstance(geo_obj, Face3D) \
                    else from_polyface3d(geo_obj)
                self.draw_brep_wires.append((rh_obj, col, 1))
            elif dis_obj.display_mode == 'Points':
                for pt in geo_obj.vertices:
                    self.draw_point.append(
                        (from_point3d(pt), rd.PointStyle.RoundSimple, 3, col))

        elif isinstance(geo_obj, (Mesh3D, Mesh2D)):
            # translate display mesh
            rh_obj = from_mesh3d(geo_obj) if isinstance(geo_obj, Mesh3D) \
                else from_mesh2d(geo_obj)
            if dis_obj.display_mode in ('Surface', 'SurfaceWithEdges'):
                mat = rd.DisplayMaterial(col, 1 - (dis_obj.color.a / 255))
                t_color = argb_color_to_color(dis_obj.color)
                rh_obj.VertexColors.CreateMonotoneMesh(t_color)
                self.draw_mesh_shaded.append((rh_obj, mat))
                if dis_obj.display_mode == 'SurfaceWithEdges':
                    self.draw_mesh_wires.append((rh_obj, black(), 1))
            elif dis_obj.display_mode == 'Wireframe':
                self.draw_mesh_wires.append((rh_obj, col, 1))
            elif dis_obj.display_mode == 'Points':
                self.draw_mesh_vertices.append((rh_obj, col))

        elif isinstance(geo_obj, Sphere):
            # translate analysis sphere
            if dis_obj.display_mode in ('Surface', 'SurfaceWithEdges'):
                rh_obj = from_sphere(geo_obj)
                mat = rd.DisplayMaterial(col, 1 - (dis_obj.color.a / 255))
                self.draw_brep_shaded.append((rh_obj.ToBrep(), mat))
                if dis_obj.display_mode == 'SurfaceWithEdges':
                    self.draw_sphere.append((rh_obj, black()))
            elif dis_obj.display_mode == 'Wireframe':
                self.draw_sphere.append((from_sphere(geo_obj), col))
            elif dis_obj.display_mode == 'Points':
                pt_arg = (from_point3d(geo_obj.center),
                          rd.PointStyle.RoundSimple, 3, col)
                self.draw_point.append(pt_arg)

        elif isinstance(geo_obj, Polygon2D):
            # translate display polygon
            lw = 1 if isinstance(dis_obj.line_width, Default) \
                else int(dis_obj.line_width)
            if dis_obj.line_type == 'Continuous':
                self.draw_curve.append((from_polygon2d(geo_obj), col, lw))
            else:  # ensure the line pattern is respected
                verts = [from_point2d(pt) for pt in geo_obj.vertices]
                self.draw_patterned_polyline.append(
                    (verts, col, self.LINE_TYPE[dis_obj.line_type], lw, True))

        elif isinstance(geo_obj, (Arc3D, Arc2D)):
            # translate display arc
            lw = 1 if isinstance(dis_obj.line_width, Default) \
                else int(dis_obj.line_width)
            if dis_obj.line_type == 'Continuous':
                crv = from_arc3d(geo_obj) if isinstance(geo_obj, Arc3D) else \
                    from_arc2d(geo_obj)
                if geo_obj.is_circle:
                    self.draw_circle.append((crv, col, lw))
                else:
                    self.draw_arc.append((crv, col, lw))
            else:  # ensure the line pattern is respected
                p_line = geo_obj.to_polyline(
                    int(abs(geo_obj.a2 - geo_obj.a1) / 0.0523599))
                if geo_obj.is_circle:
                    arc_verts, closed = p_line.vertices[:-1], True
                else:
                    arc_verts, closed = p_line.vertices, False
                verts = [from_point3d(pt) for pt in arc_verts] \
                    if isinstance(p_line, Polyline3D) else \
                    [from_point2d(pt) for pt in arc_verts]
                self.draw_patterned_polyline.append(
                    (verts, col, self.LINE_TYPE[dis_obj.line_type], lw, closed))

        elif isinstance(geo_obj, Cone):
            # translate analysis cone
            if dis_obj.display_mode in ('Surface', 'SurfaceWithEdges'):
                rh_obj = from_cone(geo_obj)
                mat = rd.DisplayMaterial(col, 1 - (dis_obj.color.a / 255))
                self.draw_brep_shaded.append((rh_obj.ToBrep(True), mat))
                if dis_obj.display_mode == 'SurfaceWithEdges':
                    self.draw_cone.append((rh_obj, black()))
            elif dis_obj.display_mode == 'Wireframe':
                self.draw_cone.append((from_cone(geo_obj), col))
            elif dis_obj.display_mode == 'Points':
                pt_arg = (from_point3d(geo_obj.vertex),
                          rd.PointStyle.RoundSimple, 3, col)
                self.draw_point.append(pt_arg)

        elif isinstance(geo_obj, Cylinder):
            # translate analysis cylinder
            if dis_obj.display_mode in ('Surface', 'SurfaceWithEdges'):
                rh_obj = from_cylinder(geo_obj)
                mat = rd.DisplayMaterial(col, 1 - (dis_obj.color.a / 255))
                self.draw_brep_shaded.append((rh_obj.ToBrep(True, True), mat))
                if dis_obj.display_mode == 'SurfaceWithEdges':
                    self.draw_cylinder.append((rh_obj, black()))
            elif dis_obj.display_mode == 'Wireframe':
                self.draw_cylinder.append((from_cylinder(geo_obj), col))
            elif dis_obj.display_mode == 'Points':
                pt_arg = (from_point3d(geo_obj.center),
                          rd.PointStyle.RoundSimple, 3, col)
                self.draw_point.append(pt_arg)

        elif isinstance(geo_obj, Plane):
            # translate analysis planes
            pln = from_plane(geo_obj)
            r = 10  # default radius for a plane object in rhino model units
            # grid lines
            if dis_obj.show_grid:
                for j in range(-5, 6):
                    i = j / 5
                    p0 = pln.PointAt(i * r, -r)
                    p1 = pln.PointAt(i * r, r)
                    self.draw_line.append((p0, p1, col, 1))
                    p0 = pln.PointAt(-r, i * r)
                    p1 = pln.PointAt(r, i * r)
                    self.draw_line.append((p0, p1, col, 1))
            # axes
            if dis_obj.show_axes:
                self.draw_line.append((pln.Origin, pln.Origin + pln.XAxis * r, col, 3))
                self.draw_line.append((pln.Origin, pln.Origin + pln.YAxis * r, col, 3))

    def sort_shaded_objects(self):
        """Sort shaded objects according to their transparency.

        This ensures that the resulting display has visible objects behind
        any transparent objects.
        """
        if len(self.draw_brep_shaded) != 0:
            trans = (args[1].Transparency for args in self.draw_brep_shaded)
            self.draw_brep_shaded = \
                [a for _, a in sorted(zip(trans, self.draw_brep_shaded))]
        if len(self.draw_mesh_shaded) != 0:
            trans = (args[1].Transparency for args in self.draw_mesh_shaded)
            self.draw_mesh_shaded = \
                [a for _, a in sorted(zip(trans, self.draw_mesh_shaded))]

    def ToString(self):
        """Overwrite .NET ToString method."""
        return 'VisualizationSet Converter'



"""Module for exporting components from Grasshopper with all metadata and source code.
"""
import json
import os


class Component(object):
    """Grasshopper component wrapper used to serialize component properties to dict.

    Args:
        name: Text for the name of the component.
        nickname: Text for the nickname of the component.
        description: Text for the description of the component.
        code: Text for all the Python code of the component.
        category: Text for all Python code in the component (including import statements)
        subcategory: Text for the subcategory of the component.
        version: Text for the version of the component formatted as a 3-number
            semantic version.
    """
    # dictionary to map plugin-specific language to generic slugs
    MAPPING = {
        'grasshopper': '{{plugin}}',
        'Grasshopper': '{{Plugin}}',
        'GH': '{{PLGN}}',
        'Food4Rhino': '{{Package_Manager}}',
        'rhino': '{{cad}}',
        'Rhino': '{{Cad}}'
    }

    def __init__(self, name, nickname, description, code,
                 category, subcategory, version):
        """Grasshopper component wrapper."""
        self.name = name.replace('\r\n', '\n')
        self.nickname = nickname.replace('\r\n', '\n')
        self.description = description.replace('\r\n', '\n')
        self.code = code.replace('\r\n', '\n')
        self.category = category.replace('\r\n', '\n')
        self.subcategory = subcategory.replace('\r\n', '\n')
        self.version = version
        self._inputs = []
        self._outputs = []

    @classmethod
    def from_gh_component(cls, component):
        """Create Component from a Grasshopper component object.

        Args:
            component: A Grasshopper component object.
        """
        comp = cls(component.Name, component.NickName, component.Description,
                   component.Code, component.Category, component.SubCategory,
                   component.Message)

        for inp in component.Params.Input:
            comp.add_input(Port.from_gh_port(inp))

        for out in component.Params.Output:
            if out.Name == 'out':
                continue
            comp.add_output(Port.from_gh_port(out))

        return comp

    @property
    def inputs(self):
        """Get a list of Port objects for the component Inputs."""
        return self._inputs

    @property
    def outputs(self):
        """Get a list of Port objects for the component Outputs."""
        return self._outputs

    def add_input(self, inp):
        """Add an input for the component.

        Args:
            inp: A Port object for the input.
        """
        assert isinstance(inp, Port)
        self._inputs.append(inp)

    def add_output(self, out):
        """Add an output for the component.

        Args:
            out: A Port object for the output.
        """
        assert isinstance(out, Port)
        self._outputs.append(out)

    def to_dict(self):
        """Get the Component instance as a dictionary."""
        component = {
            'name': self.name,
            'version': self.version,
            'nickname': self.nickname,
            'description': self.description,
            'code': self._clean_code(),
            'category': self.category,
            'subcategory': self.subcategory,
            'inputs': [i.to_dict() for i in self.inputs],
            'outputs': [[i.to_dict() for i in self.outputs]]
        }
        return component

    def to_json(self, folder, name=None, indent=2):
        """Write the Component instance to a JSON file.

        Args:
            folder: Text for the folder into which the JSON should be written.
            name: Text for the file name for the JSON.
            indent: Integer for the number of spaces in an indent.
        """
        name = name or self.name.replace(' ', '_')
        if not name.lower().endswith('.json'):
            name = '%s.json' % name
        fp = os.path.join(folder, name)
        with open(fp, 'w') as outf:
            json.dump(self.to_dict(), outf, indent=indent)

    def _clean_code(self):
        """Clean up the text of the code before export.

        This replaces plugin-specific text like "Grasshopper" with the generic slugs
        in the MAPPING property of this class.
        """
        code = self.code
        code = code.split('\n')

        for count, line in enumerate(code):
            if line.startswith('ghenv.Component.AdditionalHelpFromDocStrings'):
                break

        gist = '\n'.join(code[count + 1:])

        for o, t in self.MAPPING.items():
            gist = gist.replace(o, t)

        return gist


class Port(object):
    """Grasshopper port wrapper used to serialize inputs and outputs to dict.

    Args:
        name: Text for the name of the input or output.
        description: Text for the description of the input or output.
        default_value: Default value for the input or output.
        value_type: Text the type of input or output (eg. bool)
        access_type: Text for list vs. item access.
    """

    def __init__(self, name, description=None, default_value=None, value_type=None,
                 access_type=None):
        self.name = name.replace('\r\n', '\n')
        self.description = description.replace('\r\n', '\n')
        self.default_value = default_value
        self.value_type = value_type
        self.access_type = str(access_type)

    @classmethod
    def from_gh_port(cls, port):
        """Create Port from a Grasshopper port object.

        Args:
            port: A Grasshopper port object, typically accessed by iterating over the
                component.Params.Input or component.Params.Output properties.
        """
        if hasattr(port, 'TypeHint'):  # it's an input
            v = port.VolatileData
            if v.IsEmpty:
                value = None
            else:
                values = tuple(str(i.Value).lower() if port.TypeHint.TypeName == 'bool'
                               else i.Value for i in v.AllData(True))
                try:
                    value = tuple(v.replace('\\\\', '\\').replace('\\', '\\\\')
                                  for v in values)
                except AttributeError:
                    # non string type
                    value = values

            if value and str(port.Access) == 'item':
                value = value[0]

            return cls(port.Name, port.Description, value, port.TypeHint.TypeName,
                       str(port.Access))
        else:  # it's an output
            return cls(port.Name, port.Description, None, None, None)

    def to_dict(self):
        """Translate Port instance to a dictionary."""
        return {
            'name': self.name,
            'description': self.description,
            'default': self.default_value,
            'type': self.value_type,
            'access': self.access_type
        }



"""Functions for dealing with inputs and outputs from Grasshopper components."""
import collections
import array
import math

try:
    from System import Object
    from System.Windows import Forms
    from System import Environment
    import System.Security.Principal as sec
    import System.Threading.Tasks as tasks
except ImportError:
    print("Failed to import System.")

try:
    import Rhino.UI as rui
except ImportError as e:
    raise ImportError("Failed to import Rhino.\n{}".format(e))

try:
    import scriptcontext
except ImportError:  # No Rhino doc is available.
    print('Failed to import Rhino scriptcontext. Document counters will be unavailable.')

try:
    from Grasshopper.Kernel import GH_RuntimeMessageLevel as Message
    from Grasshopper.Kernel.Types import GH_ObjectWrapper as Goo
    from Grasshopper import DataTree
    from Grasshopper.Kernel.Data import GH_Path as Path
    from Grasshopper import Instances
except ImportError:
    raise ImportError("Failed to import Grasshopper.")


def give_warning(component, message):
    """Give a warning message (turning the component orange).

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
        message: Text string for the warning message.
    """
    component.AddRuntimeMessage(Message.Warning, message)


def give_remark(component, message):
    """Give an remark message (giving a little grey ballon in the upper left).

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
        message: Text string for the warning message.
    """
    component.AddRuntimeMessage(Message.Remark, message)


def give_popup_message(message, window_title='', icon_type='information'):
    """Give a Windows popup message with an OK button.

    Useful in cases where you really need the user to pay attention to the message.

    Args:
        message: Text string for the popup message.
        window_title: Text string for the title of the popup window. (Default: "").
        icon_type: Text for the type of icon to be used. (Default: "information").
            Choose from the following options.

            * information
            * warning
            * error

    """
    icon_types = {
        'information': Forms.MessageBoxIcon.Information,
        'warning': Forms.MessageBoxIcon.Warning,
        'error': Forms.MessageBoxIcon.Error
    }
    icon = icon_types[icon_type]
    buttons = Forms.MessageBoxButtons.OK
    rui.Dialogs.ShowMessageBox(message, window_title, buttons, icon)


def all_required_inputs(component):
    """Check that all required inputs on a component are present.

    Note that this method needs required inputs to be written in the correct
    format on the component in order to work (required inputs have a
    single _ at the start and no _ at the end).

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.

    Returns:
        True if all required inputs are present. False if they are not.
    """
    is_input_missing = False
    for param in component.Params.Input:
        if param.NickName.startswith('_') and not param.NickName.endswith('_'):
            missing = False
            if not param.VolatileDataCount:
                missing = True
            elif param.VolatileData[0][0] is None:
                missing = True

            if missing is True:
                msg = 'Input parameter {} failed to collect data!'.format(param.NickName)
                print(msg)
                give_warning(component, msg)
                is_input_missing = True
    return not is_input_missing


def is_user_admin():
    """Get a Boolean for whether the current User has Admin access through Rhino."""
    principal = sec.WindowsPrincipal(sec.WindowsIdentity.GetCurrent())
    return principal.IsInRole(sec.WindowsBuiltInRole.Administrator)


def local_processor_count():
    """Get an integer for the number of processors on this machine.

    If, for whatever reason, the number of processors could not be sensed,
    None will be returned.
    """
    return Environment.ProcessorCount


def recommended_processor_count():
    """Get an integer for the recommended number of processors for parallel calculation.

    This should be one less than the number of processors available on this machine
    unless the machine has only one processor, in which case 1 will be returned.
    If, for whatever reason, the number of processors could not be sensed, a value
    of 1 will be returned.
    """
    cpu_count = local_processor_count()
    return 1 if cpu_count is None or cpu_count <= 1 else cpu_count - 1


def run_function_in_parallel(parallel_function, object_count, cpu_count=None):
    """Run any function in parallel given a number of objects to be iterated over.

    This method can run the calculation in a manner that targets a given CPU
    count and will also run the function normally (without the use of Tasks)
    if only one CPU is specified.

    Args:
        parallel_function: A function which will be iterated over in a parallelized
            manner. This function should have a single input argument, which
            is the integer of the object to be simulated. Note that, in order
            for this function to be successfully parallelized, any lists of
            output data must be set up beforehand and this parallel_function
            should simply be replacing the data in this pre-created list.
        object_count: An integer for the number of objects which will be iterated over
            in a parallelized manner.
        cpu_count: An integer for the number of CPUs to be used in the intersection
            calculation. The ladybug_rhino.grasshopper.recommended_processor_count
            function can be used to get a recommendation. If set to None, all
            available processors will be used. (Default: None).
    """

    def compute_each_object_group(worker_i):
        """Run groups of objects so that only the cpu_count is used."""
        start_i, stop_i = obj_groups[worker_i]
        for count in range(start_i, stop_i):
            parallel_function(count)

    if cpu_count is not None and cpu_count > 1:
        # group the objects in order to meet the cpu_count
        worker_count = min((cpu_count, object_count))
        i_per_group = int(math.ceil(object_count / worker_count))
        obj_groups = [[x, x + i_per_group] for x in range(0, object_count, i_per_group)]
        obj_groups[-1][-1] = object_count  # ensure the last group ends with obj count

    if cpu_count is None:  # use all available CPUs
        tasks.Parallel.ForEach(range(object_count), parallel_function)
    elif cpu_count <= 1:  # run everything on a single processor
        for i in range(object_count):
            parallel_function(i)
    else:  # run the groups in a manner that meets the CPU count
        tasks.Parallel.ForEach(range(len(obj_groups)), compute_each_object_group)


def component_guid(component):
    """Get the unique ID associated with a specific component.

    This ID remains the same every time that the component is run.

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.

    Returns:
        Text string for the component's unique ID.
    """
    return component.GetHashCode().ToString()


def bring_to_front(component):
    """Bring a component to the front of the canvas so that it is always executed last.

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
    """
    doc = Instances.ActiveCanvas.Document.Objects
    in_front = doc[doc.Count - 1].InstanceGuid.Equals(component.InstanceGuid)
    if not in_front:  # bring the component to the top
        component.OnPingDocument().DeselectAll()  # de-select all components
        component.Attributes.Selected = True  # select the component to move
        component.OnPingDocument().BringSelectionToTop()
        component.Attributes.Selected = False  # de-select the component after moving


def send_to_back(component):
    """Send a component to the back of the canvas so that it is always executed first.

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
    """
    doc = Instances.ActiveCanvas.Document.Objects
    in_back = doc[0].InstanceGuid.Equals(component.InstanceGuid)
    if not in_back:  # send the component to the back
        component.OnPingDocument().SelectAll()  # select all components
        component.Attributes.Selected = False  # de-select the component to move
        component.OnPingDocument().BringSelectionToTop()
        component.OnPingDocument().DeselectAll()  # de-select all after moving


def wrap_output(output):
    """Wrap Python objects as Grasshopper generic objects.

    Passing output lists of Python objects through this function can greatly reduce
    the time needed to run the component since Grasshopper can spend a long time
    figuring out the object type is if it is not recognized.  However, if the number
    of output objects is usually < 100, running this method won't make a significant
    difference and so there's no need to use it.

    Args:
        output: A list of values to be wrapped as a generic Grasshopper Object (GOO).
    """
    if not output:
        return output
    try:
        return (Goo(i) for i in output)
    except Exception as e:
        raise ValueError('Failed to wrap {}:\n{}.'.format(output, e))


def objectify_output(object_name, output_data):
    """Wrap data into a single custom Python object that can later be de-serialized.

    This is meant to address the same issue as the wrap_output method but it does
    so by simply hiding the individual items from the Grasshopper UI within a custom
    parent object that other components can accept as input and de-objectify to
    get access to the data. This strategy is also useful for the case of standard
    object types like integers where the large number of data points slows down
    the Grasshopper UI when they are output.

    Args:
        object_name: Text for the name of the custom object that will wrap the data.
            This is how the object will display in the Grasshopper UI.
        output_data: A list of data to be stored under the data property of
            the output object.
    """
    class Objectifier(object):
        """Generic class for objectifying data."""

        def __init__(self, name, data):
            self.name = name
            self.data = data

        def ToString(self):
            return '{} ({} items)'.format(self.name, len(self.data))

    return Objectifier(object_name, output_data)


def de_objectify_output(objectified_data):
    """Extract the data from an object that was output from the objectify_output method.

    Args:
        objectified_data: An object that has been output from the objectify_output
            method for which data will be returned.
    """
    return objectified_data.data


def document_counter(counter_name):
    """Get an integer for a counter name that advances each time this function is called.

    Args:
        counter_name: The name of the counter that will be advanced.
    """
    try:  # get the counter and advance it one value
        scriptcontext.sticky[counter_name] += 1
    except KeyError:  # first time that the counter is called
        scriptcontext.sticky[counter_name] = 1
    return scriptcontext.sticky[counter_name]


def longest_list(values, index):
    """Get a value from a list while applying Grasshopper's longest-list logic.

    Args:
        values: An array of values from which a value will be pulled following
            longest list logic.
        index: Integer for the index of the item in the list to return. If this
            index is greater than the length of the values, the last item of the
            list will be returned.
    """
    try:
        return values[index]
    except IndexError:
        return values[-1]


def data_tree_to_list(input):
    """Convert a grasshopper DataTree to nested lists of lists.

    Args:
        input: A Grasshopper DataTree.

    Returns:
        listData -- A list of namedtuples (path, dataList)
    """
    all_data = list(range(len(input.Paths)))
    pattern = collections.namedtuple('Pattern', 'path list')

    for i, path in enumerate(input.Paths):
        data = input.Branch(path)
        branch = pattern(path, [])

        for d in data:
            if d is not None:
                branch.list.append(d)

        all_data[i] = branch

    return all_data


def list_to_data_tree(input, root_count=0, s_type=object):
    """Transform nested of lists or tuples to a Grasshopper DataTree.

    Args:
        input: A nested list of lists to be converted into a data tree.
        root_count: An integer for the starting path of the data tree.
        s_type: An optional data type (eg. float, int, str) that defines all of the
            data in the data tree. The default (object) works will all data types
            but the conversion to data trees can be more efficient if a more
            specific type is specified.
    """

    def proc(input, tree, track):
        for i, item in enumerate(input):
            if isinstance(item, (list, tuple, array.array)):  # ignore iterables like str
                track.append(i)
                proc(item, tree, track)
                track.pop()
            else:
                tree.Add(item, Path(*track))

    if input is not None:
        t = DataTree[s_type]()
        proc(input, t, [root_count])
        return t


def merge_data_tree(data_trees, s_type=object):
    """Merge a list of grasshopper DataTrees into a single DataTree.

    Args:
        input: A list Grasshopper DataTrees to be merged into one.
        s_type: An optional data type (eg. float, int, str) that defines all of the
            data in the data tree. The default (object) works will all data types
            but the conversion to data trees can be more efficient if a more
            specific type is specified.
    """
    comb_tree = DataTree[s_type]()
    for d_tree in data_trees:
        for p, branch in zip(d_tree.Paths, d_tree.Branches):
            comb_tree.AddRange(branch, p)
    return comb_tree


def flatten_data_tree(input):
    """Flatten and clean a grasshopper DataTree into a single list and a pattern.

    Args:
        input: A Grasshopper DataTree.

    Returns:
        A tuple with two elements

        -   all_data -- All data in DataTree as a flattened list.

        -   pattern -- A dictionary of patterns as namedtuple(path, index of last item
            on this path, path Count). Pattern is useful to un-flatten the list
            back to a DataTree.
    """
    Pattern = collections.namedtuple('Pattern', 'path index count')
    pattern = dict()
    all_data = []
    index = 0  # Global counter for all the data
    for i, path in enumerate(input.Paths):
        count = 0
        data = input.Branch(path)

        for d in data:
            if d is not None:
                count += 1
                index += 1
                all_data.append(d)

        pattern[i] = Pattern(path, index, count)

    return all_data, pattern


def unflatten_to_data_tree(all_data, pattern):
    """Create DataTree from a single flattened list and a pattern.

    Args:
        all_data: A flattened list of all data
        pattern: A dictionary of patterns
            Pattern = namedtuple('Pattern', 'path index count')

    Returns:
        data_tree -- A Grasshopper DataTree.
    """
    data_tree = DataTree[Object]()
    for branch in range(len(pattern)):
        path, index, count = pattern[branch]
        data_tree.AddRange(all_data[index - count:index], path)

    return data_tree


def recipe_result(result):
    """Process a recipe result and handle the case that it's a list of list.

    Args:
        result: A recipe result to be processed.
    """
    if isinstance(result, (list, tuple)):
        return list_to_data_tree(result)
    return result


def hide_output(component, output_index):
    """Hide one of the outputs of a component.

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
        output_index: Integer for the index of the output to hide.
    """
    component.Params.Output[output_index].Hidden = True


def show_output(component, output_index):
    """Show one of the outputs of a component.

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
        output_index: Integer for the index of the output to hide.
    """
    component.Params.Output[output_index].Hidden = False


def schedule_solution(component, milliseconds):
    """Schedule a new Grasshopper solution after a specified time interval.

    Args:
        component: The grasshopper component object, which can be accessed through
            the ghenv.Component call within Grasshopper API.
        milliseconds: Integer for the number of milliseconds after which the
            solution should happen.
    """
    doc = component.OnPingDocument()
    doc.ScheduleSolution(milliseconds)



"""Classes for colorized versions of various Rhino objects like points."""
from .color import black

try:
    import System.Guid as guid
except ImportError as e:
    raise ImportError('Failed to import System.\n{}'.format(e))

try:
    import Rhino as rh
except ImportError as e:
    raise ImportError('Failed to import Rhino.\n{}'.format(e))

try:
    import Grasshopper as gh
except ImportError:
    print('Failed to import Grasshopper.\nColorized objects are not available.')


class ColoredPoint(gh.Kernel.Types.GH_GeometricGoo[rh.Geometry.Point3d],
                   gh.Kernel.IGH_BakeAwareData, gh.Kernel.IGH_PreviewData):
    """A Point object with a set-able color property to change its color in Grasshopper.

    Args:
        point: A Rhino Point3d object.
    """

    def __init__(self, point):
        """Initialize ColoredPoint."""
        self.point = point
        self.color = black()

    def DuplicateGeometry(self):
        point = rh.Geometry.Point3d(self.point.X, self.point.Y, self.point.Z)
        new_pt = ColoredPoint(point)
        new_pt.color = self.color
        return new_pt

    def get_TypeName(self):
        return "Colored Point"

    def get_TypeDescription(self):
        return "Colored Point"

    def ToString(self):
        return '{}, {}, {}'.format(self.color.R, self.color.G, self.color.B)

    def Transform(self, xform):
        point = rh.Geometry.Point3d(self.point.X, self.point.Y, self.point.Z)
        point.Transform(xform)
        new_pt = ColoredPoint(point)
        new_pt.color = self.color
        return new_pt

    def Morph(self, xmorph):
        return self.DuplicateGeometry()

    def DrawViewportWires(self, args):
        args.Pipeline.DrawPoint(
            self.point, rh.Display.PointStyle.RoundSimple, 5, self.color)

    def DrawViewportMeshes(self, args):
        # Do not draw in meshing layer.
        pass

    def BakeGeometry(self, doc, att, id):
        id = guid.Empty
        if att is None:
            att = doc.CreateDefaultAttributes()
        att.ColorSource = rh.DocObjects.ObjectColorSource.ColorFromObject
        att.ObjectColor = self.color
        id = doc.Objects.AddPoint(self.point, att)
        return True, id


class ColoredPolyline(gh.Kernel.Types.GH_GeometricGoo[rh.Geometry.PolylineCurve],
                      gh.Kernel.IGH_BakeAwareData, gh.Kernel.IGH_PreviewData):
    """A PolylineCurve object with set-able color and thickness properties.

    Args:
        polyline: A Rhino PolylineCurve object.
    """

    def __init__(self, polyline):
        """Initialize ColoredPolyline."""
        self.polyline = polyline
        self.color = black()
        self.thickness = 1

    def DuplicateGeometry(self):
        polyline = rh.Geometry.PolylineCurve(self.polyline)
        new_pl = ColoredPolyline(polyline)
        new_pl.color = self.color
        new_pl.thickness = self.thickness
        return new_pl

    def get_TypeName(self):
        return "Colored Polyline"

    def get_TypeDescription(self):
        return "Colored Polyline"

    def ToString(self):
        return 'Polyline Curve'

    def Transform(self, xform):
        polyline = rh.Geometry.PolylineCurve(self.polyline)
        polyline.Transform(xform)
        new_pl = ColoredPolyline(polyline)
        new_pl.color = self.color
        new_pl.thickness = self.thickness
        return new_pl

    def Morph(self, xmorph):
        return self.DuplicateGeometry()

    def DrawViewportWires(self, args):
        args.Pipeline.DrawCurve(self.polyline, self.color, self.thickness)

    def DrawViewportMeshes(self, args):
        # Do not draw in meshing layer.
        pass

    def BakeGeometry(self, doc, att, id):
        id = guid.Empty
        if att is None:
            att = doc.CreateDefaultAttributes()
        att.ColorSource = rh.DocObjects.ObjectColorSource.ColorFromObject
        att.ObjectColor = self.color
        id = doc.Objects.AddCurve(self.polyline, att)
        return True, id


class ColoredLine(gh.Kernel.Types.GH_GeometricGoo[rh.Geometry.LineCurve],
                  gh.Kernel.IGH_BakeAwareData, gh.Kernel.IGH_PreviewData):
    """A LineCurve object with set-able color and thickness properties.

    Args:
        line: A Rhino LineCurve object.
    """

    def __init__(self, line):
        """Initialize ColoredPolyline."""
        self.line = line
        self.color = black()
        self.thickness = 1

    def DuplicateGeometry(self):
        line = rh.Geometry.LineCurve(self.line)
        new_pl = ColoredPolyline(line)
        new_pl.color = self.color
        new_pl.thickness = self.thickness
        return new_pl

    def get_TypeName(self):
        return "Colored Polyline"

    def get_TypeDescription(self):
        return "Colored Polyline"

    def ToString(self):
        return 'Polyline Curve'

    def Transform(self, xform):
        line = rh.Geometry.LineCurve(self.line)
        line.Transform(xform)
        new_pl = ColoredPolyline(line)
        new_pl.color = self.color
        new_pl.thickness = self.thickness
        return new_pl

    def Morph(self, xmorph):
        return self.DuplicateGeometry()

    def DrawViewportWires(self, args):
        args.Pipeline.DrawCurve(self.line, self.color, self.thickness)

    def DrawViewportMeshes(self, args):
        # Do not draw in meshing layer.
        pass

    def BakeGeometry(self, doc, att, id):
        id = guid.Empty
        if att is None:
            att = doc.CreateDefaultAttributes()
        att.ColorSource = rh.DocObjects.ObjectColorSource.ColorFromObject
        att.ObjectColor = self.color
        id = doc.Objects.AddCurve(self.line, att)
        return True, id



"""Functions specific to stable releases like changing component versions."""


def update_component_version(components, version, year=None):
    """Update the version number and copyright year for a list of components.

    Args:
        components: A list of Grasshopper component objects which will have
            their version updated. Typically, this should be the output of the
            place_plugin_components method from ladybug_rhino.versioning.gather.
        version: Text for the version of the components to update.
        year: Text for the copyright year to update.

    Returns:
        A list of Ladybug Tools component objects that have had their version
        and copyright year updated.
    """
    new_components = []
    for comp_obj in components:
        try:
            # get the code from inside the component
            in_code = comp_obj.Code.split("\n")
            code_length = len(in_code)
            out_code = ''

            # loop through the lines of code and replace the version + copyright
            for line_count, line in enumerate(in_code):
                # replace the mesage and copyright lines
                if line.startswith('ghenv.Component.Message'):
                    line = "ghenv.Component.Message = '{}'".format(version)
                elif line.startswith('# Copyright (c) ') and year is not None:
                    line = '# Copyright (c) {}, Ladybug Tools.'.format(year)

                # append the code to the output lines
                if line_count != code_length - 1:
                    out_code += line + "\n"
                else:
                    out_code += line

            comp_obj.Code = out_code  # replace the old code with updated code
            new_components.append(comp_obj)
        except Exception:  # not a Ladybug Tools component
            print('Failed to update version in "{}".'.format(comp_obj.Name))
    return new_components



"""Functions for importing OpenStudio into the Python environment."""
import os
import shutil
import sys

try:
    import clr
except ImportError as e:  # No .NET being used
    print('Failed to import CLR. OpenStudio SDK is unavailable.\n{}'.format(e))

try:
    from honeybee_energy.config import folders
except ImportError as e:
    print('Failed to import honeybee_energy. '
          'OpenStudio SDK is unavailable.\n{}'.format(e))


def load_osm(osm_path):
    """Load an OSM file to an OpenStudio SDK Model object in the Python environment.

    Args:
        osm_path: The path to an OSM file to be loaded an an OpenStudio Model.

    Returns:
        An OpenStudio Model object derived from the input osm_path.

    Usage:

    .. code-block:: python

        from ladybug_rhino.openstudio import load_osm

        # load an OpenStudio model from an OSM file
        osm_path = 'C:/path/to/model.osm'
        os_model = load_osm(osm_path)

        # get the space types from the model
        os_space_types = os_model.getSpaceTypes()
        for spt in os_space_types:
            print(spt)
    """
    # check that the file exists and OpenStudio is installed
    assert os.path.isfile(osm_path), 'No OSM file was found at "{}".'.format(osm_path)
    ops = import_openstudio()

    # load the model object and return it
    os_path = ops.OpenStudioUtilitiesCore.toPath(osm_path)
    osm_path_obj = ops.Path(os_path)
    exist_os_model = ops.Model.load(osm_path_obj)
    if exist_os_model.is_initialized():
        return exist_os_model.get()
    else:
        raise ValueError(
            'The file at "{}" does not appear to be an OpenStudio model.'.format(
                osm_path
            ))


def dump_osm(model, osm_path):
    """Dump an OpenStudio Model object to an OSM file.

    Args:
        model: An OpenStudio Model to be written to a file.
        osm_path: The path of the .osm file where the OpenStudio Model will be saved.

    Returns:
        The path to the .osm file as a string.

    Usage:

    .. code-block:: python

        from ladybug_rhino.openstudio import load_osm, dump_osm

        # load an OpenStudio model from an OSM file
        osm_path = 'C:/path/to/model.osm'
        model = load_osm(osm_path)

        # get all of the SetpointManagers and set their properties
        setpt_managers = model.getSetpointManagerOutdoorAirResets()
        for setpt in setpt_managers:
            setpt.setSetpointatOutdoorLowTemperature(19)
            setpt.setOutdoorLowTemperature(12)
            setpt.setSetpointatOutdoorHighTemperature(16)
            setpt.setOutdoorHighTemperature(22)

        # save the edited OSM over the original one
        osm = dump_osm(model, osm_path)
    """
    # check that the model is the correct object type
    ops = import_openstudio()
    assert isinstance(model, ops.Model), \
        'Expected OpenStudio Model. Got {}.'.format(type(model))

    # load the model object and return it
    os_path = ops.OpenStudioUtilitiesCore.toPath(osm_path)
    osm_path_obj = ops.Path(os_path)
    model.save(osm_path_obj, True)
    return osm_path


def import_openstudio():
    """Import the OpenStudio SDK into the Python environment.

    Returns:
        The OpenStudio NameSpace with all of the modules, classes and methods
        of the OpenStudio SDK.

    Usage:

    .. code-block:: python

        from ladybug_rhino.openstudio import import_openstudio, dump_osm
        OpenStudio = import_openstudio()

        # create a new OpenStudio model from scratch
        os_model = OpenStudio.Model()
        space_type = OpenStudio.SpaceType(os_model)

        # save the Model to an OSM
        osm_path = 'C:/path/to/model.osm'
        osm = dump_osm(os_model, osm_path)
    """
    try:  # first see if OpenStudio has already been loaded
        import OpenStudio
        return OpenStudio
    except ImportError:
        # check to be sure that the OpenStudio CSharp folder has been installed
        compatibility_url = 'https://github.com/ladybug-tools/lbt-grasshopper/wiki/' \
            '1.4-Compatibility-Matrix'
        in_msg = 'Download and install the version of OpenStudio listed in the ' \
            'Ladybug Tools compatibility matrix\n{}.'.format(compatibility_url)
        assert folders.openstudio_path is not None, \
            'No OpenStudio installation was found on this machine.\n{}'.format(in_msg)
        assert folders.openstudio_csharp_path is not None, \
            'No OpenStudio CSharp folder was found in the OpenStudio installation ' \
            'at:\n{}'.format(os.path.dirname(folders.openstudio_path))
        _copy_openstudio_lib()

        # add the OpenStudio DLL to the Common Language Runtime (CLR)
        os_dll = os.path.join(folders.openstudio_csharp_path, 'OpenStudio.dll')
        clr.AddReferenceToFileAndPath(os_dll)
        if folders.openstudio_csharp_path not in sys.path:
            sys.path.append(folders.openstudio_csharp_path)
        import OpenStudio
        return OpenStudio


def _copy_openstudio_lib():
    """Copy the openstudiolib.dll into the CSharp folder.

    This is a workaround that is necessary because the OpenStudio installer
    does not install the CSharp bindings correctly.
    """
    # see if the CSharp folder already has everything it needs
    dest_file = os.path.join(folders.openstudio_csharp_path, 'openstudiolib.dll')
    if os.path.isfile(dest_file):
        return None

    # if not, see if the openstudio_lib_path has the file that needs to be copied
    base_msg = 'The OpenStudio CSharp path at "{}" lacks the openstudiolib.dll'.format(
        folders.openstudio_csharp_path)
    assert os.path.isdir(folders.openstudio_lib_path), \
        '{}\nand there is no OpenStudio Lib installed.'.format(base_msg)
    src_file = os.path.join(folders.openstudio_lib_path, 'openstudiolib.dll')
    assert os.path.isfile(src_file), \
        '{}\nand this file was not found at "{}".'.format(base_msg, src_file)

    # copy the DLL if it exists
    shutil.copy(src_file, dest_file)



"""Functions for gathering connected components or all components on a canvas."""
try:
    import System.Drawing
except ImportError:
    raise ImportError("Failed to import System.")

try:
    import Grasshopper.Kernel as gh
    import Grasshopper.Instances as ghi
except ImportError:
    raise ImportError("Failed to import Grasshopper.")


# Master array of all identifiers of Ladybug Tools components
LADYBUG_TOOLS = ('LB', 'HB', 'DF', 'BF', 'Ladybug', 'Honeybee',
                 'Butterfly', 'HoneybeePlus')


def is_ladybug_tools(component):
    """Check if a component is a part of Ladybug Tools."""
    return component.Name.split(' ')[0] in LADYBUG_TOOLS or \
        component.Name.split('_')[0] in LADYBUG_TOOLS


def gather_canvas_components(component):
    """Get all of the Ladybug Tools components on the same canvas as the input component.

    This will also gather any Ladybug Tools components inside of clusters.

    Args:
        component: A Grasshopper component object. Typically, this should be the
            exporter component object, which can be accessed through the
            ghenv.Component call.

    Returns:
        A list of Ladybug Tools component objects on the same canvas as the
        input component. The input component is excluded from this list.
    """
    components = []
    document = component.OnPingDocument()
    for comp_obj in document.Objects:
        if type(comp_obj) == type(component):  # GHPython component
            if is_ladybug_tools(comp_obj):  # Ladybug Tools component
                components.append(comp_obj)
        elif type(comp_obj) == gh.Special.GH_Cluster:
            cluster_doc = comp_obj.Document("")
            if not cluster_doc:
                continue
            for cluster_obj in cluster_doc.Objects:
                if type(cluster_obj) == type(component) and \
                        is_ladybug_tools(cluster_obj):
                    if cluster_obj.Locked:
                        continue
                    components.append(cluster_obj)

    # remove the exporter component from the array
    components = tuple(comp for comp in components if
                       comp.InstanceGuid != component.InstanceGuid)

    return components


def gather_connected_components(component):
    """Get all of the GHPython components connected to the component's first input.

    Args:
        component: A Grasshopper component object. Typically, this should be the
            exporter component object, which can be accessed through the
            ghenv.Component call.

    Returns:
        A list of Ladybug Tools component objects that are connected to the component's
        first input.
    """
    param = component.Params.Input[0]  # components input
    sources = param.Sources

    if sources.Count == 0:
        # no component is connected
        yield []

    for src in sources:
        attr = src.Attributes
        if attr is None or attr.GetTopLevel is None:  # not exportable
            continue
        # collect components
        comp_obj = attr.GetTopLevel.DocObject
        if comp_obj is None:
            continue
        if type(comp_obj) != type(component):  # not a GHPython component
            continue

        yield comp_obj


def plugin_components(plugin_name, sub_category=None):
    """Get all of the components of a Ladybug Tools Plugin from the component server.

    Args:
        plugin_name: Text for the name of a particular plugin (aka. insect) to
            place components from (eg. "Ladybug", "Honeybee", "HB-Energy").
        sub_category: Text for a specific plugin sub-category (aka. tab) to
            be exported (eg. "1 :: Analyze Data"). If None, all components in
            the plugin will be places on the canvas. (Default: None).

    Returns:
        A dictionary of the plugin components with the name of the component as
        the key and the component object as the value.
    """
    # loop through the component server and find all the components in the plugin
    components = {}
    for proxy in ghi.ComponentServer.ObjectProxies:
        if proxy.Obsolete:
            continue
        category = proxy.Desc.Category
        subcategory = proxy.Desc.SubCategory
        # check to see if the component is in the plugin
        if category.strip() == plugin_name:  # if so, see if it is in the sub category
            if sub_category is None or subcategory.strip() == sub_category:
                if str(proxy.Kind) == 'CompiledObject':  # it's a .gha
                    components[proxy.Desc.Name] = proxy.CreateInstance()
                elif str(proxy.Kind) == 'UserObject':  # it's a .ghuser
                    components[proxy.Desc.Name] = \
                        gh.GH_UserObject(proxy.Location).InstantiateObject()
    return components


def place_component(component_reference, component_name, x_position=200, y_position=200,
                    hold_solution=False):
    """Place a single component on the canvas.

    Args:
        component_reference: A Grasshopper component reference object to be placed
            on the canvas.
        component_name: Text for the name of the component being placed.
        x_position: An integer for where in the X dimension of the canvas the
            components will be dropped. (Default: 200).
        y_position: An integer for where in the Y dimension of the canvas the
            components will be dropped. (Default: 200).
        hold_solution: Boolean to note whether to hold off on the solution after
            the component is dropped on the canvas. (Default: False).

    Returns:
        The Component Object for the components that have been dropped
        onto the canvas.
    """
    document = ghi.ActiveCanvas.Document  # get the Grasshopper document
    component_reference.Attributes.Pivot = System.Drawing.PointF(x_position, y_position)
    document.AddObject(component_reference, False, 0)

    # find the component object on the Grasshopper canvas using its name
    component = None
    for comp in document.Objects:
        if comp.Name == component_name:
            component = comp
            break

    # expire component solution so that all of the components don't run at once
    if hold_solution:
        try:
            component.ExpireSolution(False)
        except Exception:
            print('Failed to stop "{}" from running'.format(component_name))
    return component


def place_plugin_components(plugin_name, sub_category=None, x_position=200,
                            y_position=200):
    """Place all of the components of a specific Ladybug Tools Plugin on the canvas.

    Args:
        plugin_name: Text for the name of a particular plugin (aka. insect) to
            place components from (eg. "Ladybug", "Honeybee", "HB-Energy").
        sub_category: Text for a specific plugin sub-category (aka. tab) to
            be exported (eg. "1 :: Analyze Data"). If None, all components in
            the plugin will be places on the canvas. (Default: None).
        x_position: An integer for where in the X dimension of the canvas the
            components will be dropped. (Default: 200).
        y_position: An integer for where in the Y dimension of the canvas the
            components will be dropped. (Default: 200).

    Returns:
        A list of Component Objects for the components that have been dropped
        onto the canvas. These component objects can be used to update the
        version of the dropped component, change their category, etc.
    """
    # find all Components/UserObjects in the Grasshopper Component Server
    component_references = plugin_components(plugin_name, sub_category)

    # loop through the components and add them to the canvass
    components = []  # array to hold all of the dropped components
    for comp_name, comp_obj in component_references.items():
        # add object to document
        if comp_obj.Attributes:
            comp = place_component(comp_obj, comp_name, x_position, y_position, True)
            components.append(comp)
    return components


def remove_component(component):
    """Remove a Grasshopper component from the canvas.

    Args:
        component: The Grasshopper component object to be removed.
    """
    document = ghi.ActiveCanvas.Document  # get the Grasshopper document
    document.RemoveObject(component, False)
